import unittest

import requests_mock
import datetime

from pydantic_core import ValidationError

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Database, User, Container, Image, UserAttributes, DatabaseAccess, AccessType
from dbrepo.api.exceptions import ResponseCodeError, NotExistsError, ForbiddenError, MalformedError, AuthenticationError

from dbrepo.api.dto import ImageDate


class DatabaseUnitTest(unittest.TestCase):

    def test_get_databases_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database', json=[])
            # test
            response = RestClient().get_databases()
            self.assertEqual([], response)

    def test_get_databases_succeeds(self):
        exp = [
            Database(
                id=1,
                name='test',
                creator=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                             attributes=UserAttributes(theme='light')),
                owner=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                           attributes=UserAttributes(theme='light')),
                contact=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                             attributes=UserAttributes(theme='light')),
                created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                exchange_name='dbrepo',
                internal_name='test_abcd',
                is_public=True,
                container=Container(
                    id=1,
                    name='MariaDB Galera 11.1.3',
                    internal_name='mariadb',
                    host='data-db',
                    port=3306,
                    sidecar_host='data-db-sidecar',
                    sidecar_port=3305,
                    created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                    image=Image(
                        id=1,
                        registry='docker.io',
                        name='mariadb',
                        version='11.2.2',
                        dialect='org.hibernate.dialect.MariaDBDialect',
                        driver_class='org.mariadb.jdbc.Driver',
                        jdbc_method='mariadb',
                        default_port=3306,
                        date_formats=[ImageDate(id=1,
                                                created_at=datetime.datetime(2024, 3, 25, 18, 2, 14, 0,
                                                                             datetime.timezone.utc),
                                                example="2024-03-25 18:02:14",
                                                database_format='%Y-%c-%d %H:%i:%S',
                                                unix_format='yyyy-MM-dd HH:mm:ss',
                                                has_time=True)]
                    )
                )
            )
        ]
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_databases()
            self.assertEqual(exp, response)

    def test_get_database_succeeds(self):
        exp = Database(
            id=1,
            name='test',
            creator=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            owner=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                       attributes=UserAttributes(theme='light')),
            contact=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            container=Container(
                id=1,
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                host='data-db',
                port=3306,
                sidecar_host='data-db-sidecar',
                sidecar_port=3305,
                created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                image=Image(
                    id=1,
                    registry='docker.io',
                    name='mariadb',
                    version='11.2.2',
                    dialect='org.hibernate.dialect.MariaDBDialect',
                    driver_class='org.mariadb.jdbc.Driver',
                    jdbc_method='mariadb',
                    default_port=3306
                )
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1', json=exp.model_dump())
            # test
            response = RestClient().get_database(1)
            self.assertEqual(exp, response)

    def test_get_database_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1', status_code=404)
            # test
            try:
                response = RestClient().get_database(1)
            except NotExistsError as e:
                pass

    def test_get_database_invalid_dto_fails(self):
        try:
            exp = Database()
        except ValidationError as e:
            pass

    def test_get_database_unauthorized_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1', status_code=401)
            # test
            try:
                response = RestClient().get_database(1)
            except ResponseCodeError as e:
                pass

    def test_create_database_succeeds(self):
        exp = Database(
            id=1,
            name='test',
            creator=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            owner=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                       attributes=UserAttributes(theme='light')),
            contact=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            container=Container(
                id=1,
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                host='data-db',
                port=3306,
                sidecar_host='data-db-sidecar',
                sidecar_port=3305,
                created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                image=Image(
                    id=1,
                    registry='docker.io',
                    name='mariadb',
                    version='11.2.2',
                    dialect='org.hibernate.dialect.MariaDBDialect',
                    driver_class='org.mariadb.jdbc.Driver',
                    jdbc_method='mariadb',
                    default_port=3306
                )
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database', json=exp.model_dump(), status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_database(name='test', container_id=1, is_public=True)
            self.assertEqual(response.name, 'test')

    def test_create_database_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_database(name='test', container_id=1, is_public=True)
            except ForbiddenError as e:
                pass

    def test_create_database_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_database(name='test', container_id=1, is_public=True)
            except NotExistsError as e:
                pass

    def test_create_database_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database', status_code=404)
            # test
            try:
                response = RestClient().create_database(name='test', container_id=1, is_public=True)
            except AuthenticationError as e:
                pass

    def test_update_database_visibility_succeeds(self):
        exp = Database(
            id=1,
            name='test',
            creator=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            owner=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                       attributes=UserAttributes(theme='light')),
            contact=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            container=Container(
                id=1,
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                host='data-db',
                port=3306,
                sidecar_host='data-db-sidecar',
                sidecar_port=3305,
                created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                image=Image(
                    id=1,
                    registry='docker.io',
                    name='mariadb',
                    version='11.2.2',
                    dialect='org.hibernate.dialect.MariaDBDialect',
                    driver_class='org.mariadb.jdbc.Driver',
                    jdbc_method='mariadb',
                    default_port=3306
                )
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1', json=exp.model_dump(), status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.update_database_visibility(database_id=1, is_public=True)
            self.assertEqual(response.is_public, True)

    def test_update_database_visibility_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_database_visibility(database_id=1, is_public=True)
            except ForbiddenError:
                pass

    def test_update_database_visibility_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_database_visibility(database_id=1, is_public=True)
            except NotExistsError:
                pass

    def test_update_database_visibility_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1', status_code=404)
            # test
            try:
                response = RestClient().update_database_visibility(database_id=1, is_public=True)
            except AuthenticationError:
                pass

    def test_update_database_owner_succeeds(self):
        exp = Database(
            id=1,
            name='test',
            creator=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            owner=User(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='other',
                       attributes=UserAttributes(theme='light')),
            contact=User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                         attributes=UserAttributes(theme='light')),
            created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            container=Container(
                id=1,
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                host='data-db',
                port=3306,
                sidecar_host='data-db-sidecar',
                sidecar_port=3305,
                created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                image=Image(
                    id=1,
                    registry='docker.io',
                    name='mariadb',
                    version='11.2.2',
                    dialect='org.hibernate.dialect.MariaDBDialect',
                    driver_class='org.mariadb.jdbc.Driver',
                    jdbc_method='mariadb',
                    default_port=3306
                )
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/owner', json=exp.model_dump(), status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.update_database_owner(database_id=1, user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            self.assertEqual(response.owner.id, 'abdbf897-e599-4e5a-a3f0-7529884ea011')

    def test_update_database_owner_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/owner', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_database_owner(database_id=1,
                                                        user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except ForbiddenError:
                pass

    def test_update_database_owner_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/owner', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_database_owner(database_id=1,
                                                        user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except NotExistsError:
                pass

    def test_update_database_owner_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/owner', status_code=404)
            # test
            try:
                response = RestClient().update_database_owner(database_id=1,
                                                              user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except AuthenticationError:
                pass

    def test_get_database_access_succeeds(self):
        exp = DatabaseAccess(type=AccessType.READ,
                             created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                             user=User(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='other',
                                       attributes=UserAttributes(theme='light')))
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/access', json=exp.model_dump())
            # test
            response = RestClient().get_database_access(database_id=1)
            self.assertEqual(response, AccessType.READ)

    def test_get_database_access_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/access', status_code=403)
            # test
            try:
                response = RestClient().get_database_access(database_id=1)
            except ForbiddenError:
                pass

    def test_get_database_access_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/database/1/access', status_code=404)
            # test
            try:
                response = RestClient().get_database_access(database_id=1)
            except NotExistsError:
                pass

    def test_create_database_access_succeeds(self):
        exp = DatabaseAccess(type=AccessType.READ,
                             created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                             user=User(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='other',
                                       attributes=UserAttributes(theme='light')))
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', json=exp.model_dump(),
                      status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_database_access(database_id=1, type=AccessType.READ,
                                                     user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            self.assertEqual(response, exp.type)

    def test_create_database_access_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_database_access(database_id=1, type=AccessType.READ,
                                                         user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except MalformedError:
                pass

    def test_create_database_access_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=400)
            # test
            try:
                response = RestClient().create_database_access(database_id=1, type=AccessType.READ,
                                                               user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except AuthenticationError:
                pass

    def test_create_database_access_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_database_access(database_id=1, type=AccessType.READ,
                                                         user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except ForbiddenError:
                pass

    def test_create_database_access_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_database_access(database_id=1, type=AccessType.READ,
                                                         user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except NotExistsError:
                pass

    def test_update_database_access_succeeds(self):
        exp = DatabaseAccess(type=AccessType.READ,
                             created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                             user=User(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='other',
                                       attributes=UserAttributes(theme='light')))
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', json=exp.model_dump(),
                     status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.update_database_access(database_id=1, type=AccessType.READ,
                                                     user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            self.assertEqual(response, exp.type)

    def test_update_database_access_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_database_access(database_id=1, type=AccessType.READ,
                                                         user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except MalformedError:
                pass

    def test_update_database_access_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_database_access(database_id=1, type=AccessType.READ,
                                                         user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except ForbiddenError:
                pass

    def test_update_database_access_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_database_access(database_id=1, type=AccessType.READ,
                                                         user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except NotExistsError:
                pass

    def test_update_database_access_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=404)
            # test
            try:
                response = RestClient().update_database_access(database_id=1, type=AccessType.READ,
                                                               user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except AuthenticationError:
                pass

    def test_delete_database_access_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=202)
            # test
            client = RestClient(username="a", password="b")
            client.delete_database_access(database_id=1, user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')

    def test_delete_database_access_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_database_access(database_id=1, user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except MalformedError:
                pass

    def test_delete_database_access_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_database_access(database_id=1, user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except ForbiddenError:
                pass

    def test_delete_database_access_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.delete_database_access(database_id=1, user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except NotExistsError:
                pass

    def test_delete_database_access_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete('/api/database/1/access/abdbf897-e599-4e5a-a3f0-7529884ea011', status_code=404)
            # test
            try:
                RestClient().delete_database_access(database_id=1, user_id='abdbf897-e599-4e5a-a3f0-7529884ea011')
            except AuthenticationError:
                pass


if __name__ == "__main__":
    unittest.main()



import dataclasses
import os
import pika
import sys
import json
import logging

from dbrepo.api.dto import CreateData


class AmqpClient:
    """
    The AmqpClient class for communicating with the DBRepo AMQP API to import data. All parameters can be set also \
    via environment variables, e.g. set endpoint with DBREPO_ENDPOINT. You can override the constructor parameters \
    with the environment variables.

    :param broker_host: The AMQP API host. Optional. Default: "broker-service"
    :param broker_port: The AMQP API port. Optional. Default: 5672
    :param broker_virtual_host: The AMQP API virtual host. Optional. Default: "/"
    :param username: The AMQP API username. Optional.
    :param password: The AMQP API password. Optional.
    """
    broker_host: str = None
    broker_port: int = 5672
    broker_virtual_host: str = None
    username: str = None
    password: str = None

    def __init__(self,
                 broker_host: str = 'broker-service',
                 broker_port: int = 5672,
                 broker_virtual_host: str = '/',
                 username: str = None,
                 password: str = None) -> None:
        logging.basicConfig(format='%(asctime)s %(name)-12s %(levelname)-6s %(message)s', level=logging.DEBUG,
                            stream=sys.stdout)
        self.broker_host = os.environ.get('AMQP_API_HOST', broker_host)
        self.broker_port = os.environ.get('AMQP_API_PORT', broker_port)
        if os.environ.get('AMQP_API_VIRTUAL_HOST') is not None:
            self.broker_virtual_host = os.environ.get('AMQP_API_VIRTUAL_HOST')
        else:
            self.broker_virtual_host = broker_virtual_host
        self.username = os.environ.get('AMQP_API_USERNAME', username)
        self.password = os.environ.get('AMQP_API_PASSWORD', password)

    def publish(self, exchange: str, routing_key: str, data=dict) -> None:
        """
        Publishes data to a given exchange with the given routing key with a blocking connection.

        :param exchange: The exchange name.
        :param routing_key: The routing key.
        :param data: The data.
        """
        parameters = pika.ConnectionParameters(host=self.broker_host, port=self.broker_port,
                                               virtual_host=self.broker_virtual_host,
                                               credentials=pika.credentials.PlainCredentials(self.username,
                                                                                             self.password))
        connection = pika.BlockingConnection(parameters)
        channel = connection.channel()
        channel.basic_publish(exchange=exchange, routing_key=routing_key,
                              body=json.dumps(dataclasses.asdict(CreateData(data=data))))
        connection.close()



from __future__ import annotations

from dataclasses import field
from enum import Enum
import datetime
from typing import List, Optional, Any, Annotated
from pydantic import BaseModel, ConfigDict, PlainSerializer, Field

Timestamp = Annotated[
    datetime.datetime, PlainSerializer(lambda v: v.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z', return_type=str)
]


class ImageDate(BaseModel):
    id: int
    database_format: str
    unix_format: str
    has_time: bool
    created_at: Timestamp


class JwtAuth(BaseModel):
    access_token: str
    refresh_token: str
    id_token: str
    expires_in: int
    refresh_expires_in: int
    not_before_policy: int = Field(alias='not-before-policy')
    scope: str
    session_state: str
    token_type: str


class Image(BaseModel):
    id: int
    registry: str
    name: str
    version: str
    dialect: str
    driver_class: str
    jdbc_method: str
    default_port: int
    date_formats: Optional[List[ImageDate]] = field(default_factory=list)


class ImageBrief(BaseModel):
    id: int
    name: str
    version: str
    jdbc_method: str


class CreateDatabase(BaseModel):
    name: str
    container_id: int
    is_public: bool


class CreateUser(BaseModel):
    username: str
    email: str
    password: str


class UpdateUser(BaseModel):
    theme: str
    language: str
    firstname: Optional[str] = None
    lastname: Optional[str] = None
    affiliation: Optional[str] = None
    orcid: Optional[str] = None


class UserBrief(BaseModel):
    id: str
    username: str
    name: Optional[str] = None
    orcid: Optional[str] = None
    qualified_name: Optional[str] = None
    given_name: Optional[str] = None
    family_name: Optional[str] = None


class Container(BaseModel):
    id: int
    name: str
    internal_name: str
    host: str
    port: int
    image: Image
    created: Timestamp
    sidecar_host: Optional[str] = None
    sidecar_port: Optional[int] = None
    ui_host: Optional[str] = None
    ui_port: Optional[int] = None


class ContainerBrief(BaseModel):
    id: int
    name: str
    image: ImageBrief
    created: Timestamp
    internal_name: str
    running: Optional[bool] = None
    hash: Optional[str] = None


class ColumnBrief(BaseModel):
    id: int
    name: str
    alias: str
    database_id: int
    table_id: int
    internal_name: str
    column_type: ColumnType


class TableBrief(BaseModel):
    id: int
    database_id: int
    name: str
    description: Optional[str]
    internal_name: str
    is_versioned: bool
    owner: UserBrief


class UserAttributes(BaseModel):
    theme: str
    orcid: Optional[str] = None
    affiliation: Optional[str] = None


class User(BaseModel):
    id: str
    username: str
    attributes: UserAttributes
    qualified_name: Optional[str] = None
    given_name: Optional[str] = None
    family_name: Optional[str] = None
    name: Optional[str] = None


class UpdateUserTheme(BaseModel):
    theme: str


class UpdateUserPassword(BaseModel):
    password: str


class AccessType(str, Enum):
    """
    Enumeration of database access.
    """
    READ = "read"
    """The user can read all data."""

    WRITE_OWN = "write_own"
    """The user can write into self-owned tables and read all data."""

    WRITE_ALL = "write_all"
    """The user can write in all tables and read all data."""


class ColumnType(str, Enum):
    """
    Enumeration of table column data types.
    """
    CHAR = "char"
    VARCHAR = "varchar"
    BINARY = "binary"
    VARBINARY = "varbinary"
    TINYBLOB = "tinyblob"
    TINYTEXT = "tinytext"
    TEXT = "text"
    BLOB = "blob"
    MEDIUMTEXT = "mediumtext"
    MEDIUMBLOB = "mediumblob"
    LONGTEXT = "longtext"
    LONGBLOB = "longblob"
    ENUM = "enum"
    SET = "set"
    BIT = "bit"
    TINYINT = "tinyint"
    BOOL = "bool"
    SMALLINT = "smallint"
    MEDIUMINT = "mediumint"
    INT = "int"
    BIGINT = "bigint"
    FLOAT = "float"
    DOUBLE = "double"
    DECIMAL = "decimal"
    DATE = "date"
    DATETIME = "datetime"
    TIMESTAMP = "timestamp"
    TIME = "time"
    YEAR = "year"


class Language(str, Enum):
    """
    Enumeration of languages.
    """
    AB = "ab"
    AA = "aa"
    AF = "af"
    AK = "ak"
    SQ = "sq"
    AM = "am"
    AR = "ar"
    AN = "an"
    HY = "hy"
    AS = "as"
    AV = "av"
    AE = "ae"
    AY = "ay"
    AZ = "az"
    BM = "bm"
    BA = "ba"
    EU = "eu"
    BE = "be"
    BN = "bn"
    BH = "bh"
    BI = "bi"
    BS = "bs"
    BR = "br"
    BG = "bg"
    MY = "my"
    CA = "ca"
    KM = "km"
    CH = "ch"
    CE = "ce"
    NY = "ny"
    ZH = "zh"
    CU = "cu"
    CV = "cv"
    KW = "kw"
    CO = "co"
    CR = "cr"
    HR = "hr"
    CS = "cs"
    DA = "da"
    DV = "dv"
    NL = "nl"
    DZ = "dz"
    EN = "en"
    EO = "eo"
    ET = "et"
    EE = "ee"
    FO = "fo"
    FJ = "fj"
    FI = "fi"
    FR = "fr"
    FF = "ff"
    GD = "gd"
    GL = "gl"
    LG = "lg"
    KA = "ka"
    DE = "de"
    KI = "ki"
    EL = "el"
    KL = "kl"
    GN = "gn"
    GU = "gu"
    HT = "ht"
    HA = "ha"
    HE = "he"
    HZ = "hz"
    HI = "hi"
    HO = "ho"
    HU = "hu"
    IS = "is"
    IO = "io"
    IG = "ig"
    ID = "id"
    IA = "ia"
    IE = "ie"
    IU = "iu"
    IK = "ik"
    GA = "ga"
    IT = "it"
    JA = "ja"
    JV = "jv"
    KN = "kn"
    KR = "kr"
    KS = "ks"
    KK = "kk"
    RW = "rw"
    KV = "kv"
    KG = "kg"
    KO = "ko"
    KJ = "kj"
    KU = "ku"
    KY = "ky"
    LO = "lo"
    LA = "la"
    LV = "lv"
    LB = "lb"
    LI = "li"
    LN = "ln"
    LT = "lt"
    LU = "lu"
    MK = "mk"
    MG = "mg"
    MS = "ms"
    ML = "ml"
    MT = "mt"
    GV = "gv"
    MI = "mi"
    MR = "mr"
    MH = "mh"
    RO = "ro"
    MN = "mn"
    NA = "na"
    NV = "nv"
    ND = "nd"
    NG = "ng"
    NE = "ne"
    SE = "se"
    NO = "no"
    NB = "nb"
    NN = "nn"
    II = "ii"
    OC = "oc"
    OJ = "oj"
    OR = "or"
    OM = "om"
    OS = "os"
    PI = "pi"
    PA = "pa"
    PS = "ps"
    FA = "fa"
    PL = "pl"
    PT = "pt"
    QU = "qu"
    RM = "rm"
    RN = "rn"
    RU = "ru"
    SM = "sm"
    SG = "sg"
    SA = "sa"
    SC = "sc"
    SR = "sr"
    SN = "sn"
    SD = "sd"
    SI = "si"
    SK = "sk"
    SL = "sl"
    SO = "so"
    ST = "st"
    NR = "nr"
    ES = "es"
    SU = "su"
    SW = "sw"
    SS = "ss"
    SV = "sv"
    TL = "tl"
    TY = "ty"
    TG = "tg"
    TA = "ta"
    TT = "tt"
    TE = "te"
    TH = "th"
    BO = "bo"
    TI = "ti"
    TO = "to"
    TS = "ts"
    TN = "tn"
    TR = "tr"
    TK = "tk"
    TW = "tw"
    UG = "ug"
    UK = "uk"
    UR = "ur"
    UZ = "uz"
    VE = "ve"
    VI = "vi"
    VO = "vo"
    WA = "wa"
    CY = "cy"
    FY = "fy"
    WO = "wo"
    XH = "xh"
    YI = "yi"
    YO = "yo"
    ZA = "za"
    ZU = "zu"


class DatabaseAccess(BaseModel):
    type: AccessType
    user: User
    created: Timestamp


class CreateAccess(BaseModel):
    type: AccessType


class UpdateAccess(BaseModel):
    type: AccessType


class IdentifierTitle(BaseModel):
    """
    Title of an identifier. See external documentation: https://support.datacite.org/docs/datacite-metadata-schema-v44-mandatory-properties#3-title.
    """
    id: int
    title: str
    language: Optional[Language] = None
    type: Optional[TitleType] = None


class CreateIdentifierTitle(BaseModel):
    title: str
    language: Optional[Language] = None
    type: Optional[TitleType] = None


class IdentifierDescription(BaseModel):
    id: int
    description: str
    language: Optional[Language] = None
    type: Optional[DescriptionType] = None


class CreateIdentifierDescription(BaseModel):
    description: str
    language: Optional[Language] = None
    type: Optional[DescriptionType] = None


class IdentifierFunder(BaseModel):
    id: int
    funder_name: str
    funder_identifier: Optional[str] = None
    funder_identifier_type: Optional[str] = None
    scheme_uri: Optional[str] = None
    award_number: Optional[str] = None
    award_title: Optional[str] = None


class CreateIdentifierFunder(BaseModel):
    funder_name: str
    funder_identifier: Optional[str] = None
    funder_identifier_type: Optional[str] = None
    scheme_uri: Optional[str] = None
    award_number: Optional[str] = None
    award_title: Optional[str] = None


class License(BaseModel):
    identifier: str
    uri: str
    description: str


class CreateData(BaseModel):
    data: dict


class UpdateData(BaseModel):
    data: dict
    keys: dict


class DeleteData(BaseModel):
    keys: dict


class Import(BaseModel):
    location: str
    separator: str
    quote: Optional[str] = None
    skip_lines: Optional[int] = None
    false_element: Optional[bool] = None
    true_element: Optional[bool] = None
    null_element: Optional[str] = None
    line_termination: Optional[str] = None


class UpdateColumn(BaseModel):
    concept_uri: Optional[str] = None
    unit_uri: Optional[str] = None


class ModifyVisibility(BaseModel):
    is_public: bool


class ModifyOwner(BaseModel):
    id: str


class CreateTable(BaseModel):
    name: str
    constraints: CreateTableConstraints
    columns: List[CreateTableColumn] = field(default_factory=list)
    description: Optional[str] = None


class CreateTableColumn(BaseModel):
    name: str
    type: ColumnType
    null_allowed: bool
    concept_uri: Optional[str] = None
    unit_uri: Optional[str] = None
    index_length: Optional[int] = None
    size: Optional[int] = None
    d: Optional[int] = None
    dfid: Optional[int] = None
    enums: Optional[List[str]] = None
    sets: Optional[List[str]] = None


class CreateTableConstraints(BaseModel):
    uniques: List[List[str]] = field(default_factory=list)
    checks: List[str] = field(default_factory=list)
    primary_key: List[str] = field(default_factory=list)
    foreign_keys: List[CreateForeignKey] = field(default_factory=list)


class IdentifierCreator(BaseModel):
    id: int
    creator_name: str
    firstname: Optional[str] = None
    lastname: Optional[str] = None
    affiliation: Optional[str] = None
    name_type: Optional[str] = None
    name_identifier: Optional[str] = None
    name_identifier_scheme: Optional[str] = None
    name_identifier_scheme_uri: Optional[str] = None
    affiliation_identifier: Optional[str] = None
    affiliation_identifier_scheme: Optional[str] = None
    affiliation_identifier_scheme_uri: Optional[str] = None


class CreateIdentifierCreator(BaseModel):
    creator_name: str
    firstname: Optional[str] = None
    lastname: Optional[str] = None
    affiliation: Optional[str] = None
    name_type: Optional[str] = None
    name_identifier: Optional[str] = None
    name_identifier_scheme: Optional[str] = None
    name_identifier_scheme_uri: Optional[str] = None
    affiliation_identifier: Optional[str] = None
    affiliation_identifier_scheme: Optional[str] = None
    affiliation_identifier_scheme_uri: Optional[str] = None


class RelatedIdentifier(BaseModel):
    id: int
    value: str
    type: RelatedIdentifierType
    relation: RelatedIdentifierRelation


class CreateRelatedIdentifier(BaseModel):
    value: str
    type: RelatedIdentifierType
    relation: RelatedIdentifierRelation


class CreateIdentifier(BaseModel):
    database_id: int
    type: IdentifierType
    creators: List[CreateIdentifierCreator]
    publication_year: int
    publisher: str
    titles: List[CreateIdentifierTitle]
    descriptions: List[CreateIdentifierDescription]
    funders: Optional[List[CreateIdentifierFunder]] = field(default_factory=list)
    doi: Optional[str] = None
    language: Optional[str] = None
    licenses: Optional[List[License]] = field(default_factory=list)
    query_id: Optional[int] = None
    table_id: Optional[int] = None
    view_id: Optional[int] = None
    query: Optional[str] = None
    query_normalized: Optional[str] = None
    execution: Optional[str] = None
    related_identifiers: Optional[List[CreateRelatedIdentifier]] = field(default_factory=list)
    result_hash: Optional[str] = None
    result_number: Optional[int] = None
    publication_day: Optional[int] = None
    publication_month: Optional[int] = None


class Identifier(BaseModel):
    id: int
    database_id: int
    type: IdentifierType
    creator: UserBrief
    status: IdentifierStatusType
    created: Timestamp
    last_modified: Timestamp
    publication_year: int
    publisher: str
    creators: List[IdentifierCreator]
    titles: List[IdentifierTitle]
    descriptions: List[IdentifierDescription]
    funders: Optional[List[IdentifierFunder]] = field(default_factory=list)
    doi: Optional[str] = None
    language: Optional[str] = None
    licenses: Optional[List[License]] = field(default_factory=list)
    query_id: Optional[int] = None
    table_id: Optional[int] = None
    view_id: Optional[int] = None
    query: Optional[str] = None
    query_normalized: Optional[str] = None
    execution: Optional[str] = None
    related_identifiers: Optional[List[RelatedIdentifier]] = field(default_factory=list)
    result_hash: Optional[str] = None
    result_number: Optional[int] = None
    publication_day: Optional[int] = None
    publication_month: Optional[int] = None


class View(BaseModel):
    id: int
    database_id: int
    name: str
    query: str
    query_hash: str
    created: Timestamp
    creator: User
    internal_name: str
    is_public: bool
    initial_view: bool
    last_modified: Timestamp
    columns: List[ViewColumn]
    identifiers: List[Identifier] = field(default_factory=list)


class CreateView(BaseModel):
    name: str
    query: str
    is_public: bool


class Result(BaseModel):
    result: Any
    headers: Any
    id: Optional[int] = None


class ViewBrief(BaseModel):
    id: int
    database_id: int
    name: str
    identifier: List[Identifier]
    query: str
    query_hash: str
    created: Timestamp
    creator: User
    internal_name: str
    is_public: bool
    initial_view: bool
    last_modified: Timestamp


class Concept(BaseModel):
    id: int
    uri: str
    created: Timestamp
    name: Optional[str] = None
    description: Optional[str] = None


class DatatypeAnalysis(BaseModel):
    separator: str
    columns: dict[str, ColumnType]
    line_termination: Optional[str] = None


class KeyAnalysis(BaseModel):
    keys: dict[str, int]


class ColumnStatistic(BaseModel):
    val_min: float
    val_max: float
    mean: float
    median: float
    std_dev: float


class ApiError(BaseModel):
    status: str
    message: str
    code: str


class TableStatistics(BaseModel):
    columns: dict[str, ColumnStatistic]


class Unit(BaseModel):
    id: int
    uri: str
    created: Timestamp
    name: Optional[str] = None
    description: Optional[str] = None


class ExecuteQuery(BaseModel):
    statement: str
    timestamp: Timestamp


class TitleType(str, Enum):
    """
    Enumeration of identifier title types.
    """
    ALTERNATIVE_TITLE = "AlternativeTitle"
    SUBTITLE = "Subtitle"
    TRANSLATED_TITLE = "TranslatedTitle"
    OTHER = "Other"


class RelatedIdentifierType(str, Enum):
    """
    Enumeration of related identifier types.
    """
    DOI = "DOI"
    URL = "URL"
    URN = "URN"
    ARK = "ARK"
    ARXIV = "arXiv"
    BIBCODE = "bibcode"
    EAN13 = "EAN13"
    EISSN = "EISSN"
    HANDLE = "Handle"
    IGSN = "IGSN"
    ISBN = "ISBN"
    ISTC = "ISTC"
    LISSN = "LISSN"
    LSID = "LSID"
    PMID = "PMID"
    PURL = "PURL"
    UPC = "UPC"
    W3ID = "w3id"


class RelatedIdentifierRelation(str, Enum):
    """
    Enumeration of related identifier types.
    """
    IS_CITED_BY = "IsCitedBy"
    CITES = "Cites"
    IS_SUPPLEMENT_TO = "IsSupplementTo"
    IS_SUPPLEMENTED_BY = "IsSupplementedBy"
    IS_CONTINUED_BY = "IsContinuedBy"
    CONTINUES = "Continues"
    IS_DESCRIBED_BY = "IsDescribedBy"
    DESCRIBES = "Describes"
    HAS_METADATA = "HasMetadata"
    IS_METADATA_FOR = "IsMetadataFor"
    HAS_VERSION = "HasVersion"
    IS_VERSION_OF = "IsVersionOf"
    IS_NEW_VERSION_OF = "IsNewVersionOf"
    IS_PREVIOUS_VERSION_OF = "IsPreviousVersionOf"
    IS_PART_OF = "IsPartOf"
    HAS_PART = "HasPart"
    IS_PUBLISHED_IN = "IsPublishedIn"
    IS_REFERENCED_BY = "IsReferencedBy"
    REFERENCES = "References"
    IS_DOCUMENTED_BY = "IsDocumentedBy"
    DOCUMENTS = "Documents"
    IS_COMPILED_BY = "IsCompiledBy"
    COMPILES = "Compiles"
    IS_VARIANT_FORM_OF = "IsVariantFormOf"
    IS_ORIGINAL_FORM_OF = "IsOriginalFormOf"
    IS_IDENTICAL_TO = "IsIdenticalTo"
    IS_REVIEWED_BY = "IsReviewedBy"
    REVIEWS = "Reviews"
    IS_DERIVED_FROM = "IsDerivedFrom"
    IS_SOURCE_OF = "IsSourceOf"
    IS_REQUIRED_BY = "IsRequiredBy"
    REQUIRES = "Requires"
    IS_OBSOLETED_BY = "IsObsoletedBy"
    OBSOLETES = "Obsoletes"


class DescriptionType(str, Enum):
    """
    Enumeration of identifier description types.
    """
    ABSTRACT = "Abstract"
    METHODS = "Methods"
    SERIES_INFORMATION = "SeriesInformation"
    TABLE_OF_CONTENTS = "TableOfContents"
    TECHNICAL_INFO = "TechnicalInfo"
    OTHER = "Other"


class QueryType(str, Enum):
    """
    Enumeration of query types.
    """
    VIEW = "view"
    """The query was executed as part of a view."""

    QUERY = "query"
    """The query was executed as subset."""


class IdentifierType(str, Enum):
    """
    Enumeration of identifier types.
    """
    VIEW = "view"
    """The identifier is identifying a view."""

    SUBSET = "subset"
    """The identifier is identifying a subset."""

    DATABASE = "database"
    """The identifier is identifying a database."""

    TABLE = "table"
    """The identifier is identifying a table."""


class IdentifierStatusType(str, Enum):
    """
    Enumeration of identifier status types.
    """
    PUBLISHED = "published"
    """The identifier is published and immutable."""

    DRAFT = "draft"
    """The identifier is a draft and can still be edited."""


class IdentifierType(str, Enum):
    """
    Enumeration of identifier types.
    """
    TABLE = "table"
    """The identifier identifies a table."""

    DATABASE = "database"
    """The identifier identifies a database."""

    VIEW = "view"
    """The identifier identifies a view."""

    SUBSET = "subset"
    """The identifier identifies a subset."""


class Query(BaseModel):
    id: int
    creator: User
    execution: Timestamp
    query: str
    type: QueryType
    created: Timestamp
    database_id: int
    query_hash: str
    is_persisted: bool
    result_hash: str
    query_normalized: str
    last_modified: Timestamp
    result_number: Optional[int] = None
    identifiers: List[Identifier] = field(default_factory=list)


class UpdateQuery(BaseModel):
    persist: bool


class Column(BaseModel):
    id: int
    name: str
    database_id: int
    table_id: int
    internal_name: str
    auto_generated: bool
    column_type: ColumnType
    is_public: bool
    is_null_allowed: bool
    alias: Optional[str] = None
    description: Optional[str] = None
    size: Optional[int] = None
    d: Optional[int] = None
    mean: Optional[float] = None
    median: Optional[float] = None
    concept: Optional[Concept] = None
    unit: Optional[Unit] = None
    enums: Optional[List[str]] = field(default_factory=list)
    sets: Optional[List[str]] = field(default_factory=list)
    date_format: Optional[ImageDate] = None
    index_length: Optional[int] = None
    length: Optional[int] = None
    data_length: Optional[int] = None
    max_data_length: Optional[int] = None
    num_rows: Optional[int] = None
    val_min: Optional[float] = None
    val_max: Optional[float] = None
    std_dev: Optional[float] = None


class ViewColumn(BaseModel):
    id: int
    name: str
    database_id: int
    internal_name: str
    auto_generated: bool
    column_type: ColumnType
    is_public: bool
    is_null_allowed: bool
    alias: Optional[str] = None
    size: Optional[int] = None
    d: Optional[int] = None
    mean: Optional[float] = None
    median: Optional[float] = None
    concept: Optional[Concept] = None
    unit: Optional[Unit] = None
    date_format: Optional[ImageDate] = None
    index_length: Optional[int] = None
    length: Optional[int] = None


class Table(BaseModel):
    id: int
    database_id: int
    name: str
    creator: User
    owner: User
    created: Timestamp
    columns: List[Column]
    constraints: Constraints
    internal_name: str
    is_versioned: bool
    created_by: str
    queue_name: str
    routing_key: str
    is_public: bool
    identifiers: Optional[List[Identifier]] = field(default_factory=list)
    description: Optional[str] = None
    queue_type: Optional[str] = None
    num_rows: Optional[int] = None
    data_length: Optional[int] = None
    max_data_length: Optional[int] = None
    avg_row_length: Optional[int] = None


class TableMinimal(BaseModel):
    id: int
    database_id: int


class ColumnMinimal(BaseModel):
    id: int
    table_id: int
    database_id: int


class Database(BaseModel):
    id: int
    name: str
    creator: User
    owner: User
    contact: User
    created: Timestamp
    exchange_name: str
    internal_name: str
    is_public: bool
    container: Container
    identifiers: Optional[List[Identifier]] = field(default_factory=list)
    subsets: Optional[List[Identifier]] = field(default_factory=list)
    description: Optional[str] = None
    tables: Optional[List[Table]] = field(default_factory=list)
    views: Optional[List[View]] = field(default_factory=list)
    image: Optional[str] = None
    accesses: Optional[List[DatabaseAccess]] = field(default_factory=list)
    exchange_type: Optional[str] = None


class Unique(BaseModel):
    id: int
    table: TableMinimal
    columns: List[ColumnMinimal]


class ForeignKeyReference(BaseModel):
    id: int
    foreign_key: ForeignKeyMinimal
    column: ColumnMinimal
    referenced_column: ColumnMinimal


class ReferenceType(str, Enum):
    """
    Enumeration of reference types.
    """
    RESTRICT = "restrict"
    CASCADE = "cascade"
    SET_NULL = "set_null"
    NO_ACTION = "no_action"
    SET_DEFAULT = "set_default"


class ForeignKeyMinimal(BaseModel):
    id: int


class ForeignKey(BaseModel):
    id: int
    name: str
    references: List[ForeignKeyReference]
    table: TableMinimal
    referenced_table: TableMinimal
    on_update: Optional[ReferenceType] = None
    on_delete: Optional[ReferenceType] = None


class CreateForeignKey(BaseModel):
    columns: List[str]
    referenced_table: str
    referenced_columns: List[str]
    on_update: Optional[ReferenceType] = None
    on_delete: Optional[ReferenceType] = None


class PrimaryKey(BaseModel):
    id: int
    table: TableMinimal
    column: ColumnMinimal


class Constraints(BaseModel):
    uniques: List[Unique]
    foreign_keys: List[ForeignKey]
    checks: List[str]
    primary_key: List[PrimaryKey]



class ResponseCodeError(Exception):
    """
    The response code is different from the expected one.
    """
    pass


class UsernameExistsError(Exception):
    """
    The username is already in use by another user.
    """
    pass


class EmailExistsError(Exception):
    """
    The e-mail address is already in use by another user.
    """
    pass


class NameExistsError(Exception):
    """
    The resource with this name exists.
    """
    pass


class NotExistsError(Exception):
    """
    The resource was not found.
    """
    pass


class ForbiddenError(Exception):
    """
    The action is not allows.
    """
    pass


class MalformedError(Exception):
    """
    The data is malformed.
    """
    pass


class QueryStoreError(Exception):
    """
    The data is malformed.
    """
    pass


class MetadataConsistencyError(Exception):
    """
    The service expected metadata that is not present.
    """
    pass


class ExternalSystemError(Exception):
    """
    The service could not communicate with the external system.
    """
    pass


class AuthenticationError(Exception):
    """
    The action requires authentication.
    """
    pass


class UploadError(Exception):
    """
    The upload was not successful.
    """
    pass



import logging
import os
import re
import sys
from tusclient import client


class UploadClient:
    """
    The UploadClient class for communicating with the DBRepo REST API. All parameters can be set also via environment \
    variables, e.g. set endpoint with DBREPO_ENDPOINT, username with DBREPO_USERNAME, etc. You can override the \
    constructor parameters with the environment variables.

    :param endpoint: The REST API endpoint. Optional. Default: "http://gateway-service/api/upload/files"
    """
    endpoint: str = None

    def __init__(self, endpoint: str = 'http://gateway-service/api/upload/files') -> None:
        logging.basicConfig(format='%(asctime)s %(name)-12s %(levelname)-6s %(message)s', level=logging.DEBUG,
                            stream=sys.stdout)
        self.endpoint = os.environ.get('REST_UPLOAD_ENDPOINT', endpoint)

    def upload(self, file_path: str) -> str:
        """
        Imports a file through the Upload Service into the Storage Service.

        :param file_path: The file path on the local machine.

        :returns: Filename on the Storage Service, if successful.
        """
        tus_client = client.TusClient(url=self.endpoint)
        uploader = tus_client.uploader(file_path=file_path)
        uploader.upload()
        m = re.search('\\/([a-f0-9]+)\\+', uploader.url)
        filename = m.group(0)[1:-1]
        logging.debug(f'uploaded file {file_path} to storage service with key: {filename}')
        return filename



import unittest

import requests_mock
import datetime

from dbrepo.RestClient import RestClient

from dbrepo.api.dto import Identifier, IdentifierType, CreateIdentifierTitle, CreateIdentifierCreator, \
    IdentifierCreator, IdentifierTitle, IdentifierDescription, CreateIdentifierDescription, Language, \
    CreateIdentifierFunder, CreateRelatedIdentifier, RelatedIdentifierRelation, RelatedIdentifierType, IdentifierFunder, \
    RelatedIdentifier, UserBrief, IdentifierStatusType
from dbrepo.api.exceptions import MalformedError, ForbiddenError, NotExistsError, ExternalSystemError, \
    AuthenticationError


class IdentifierUnitTest(unittest.TestCase):

    def test_create_identifier_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id=10,
                             database_id=1,
                             view_id=32,
                             publication_year=2024,
                             publisher='TU Wien',
                             created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                             last_modified=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             descriptions=[IdentifierDescription(id=2, description='Test Description')],
                             titles=[IdentifierTitle(id=3, title='Test Title')],
                             funders=[IdentifierFunder(id=4, funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id=7, value='10.12345/abc', relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[IdentifierCreator(id=5, creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             creator=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.post('/api/identifier', json=exp.model_dump(), status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_identifier(
                database_id=1, type=IdentifierType.VIEW,
                titles=[CreateIdentifierTitle(title='Test Title')],
                publisher='TU Wien', publication_year=2024,
                language=Language.EN,
                funders=[CreateIdentifierFunder(funder_name='FWF')],
                related_identifiers=[CreateRelatedIdentifier(value='10.12345/abc',
                                                             relation=RelatedIdentifierRelation.CITES,
                                                             type=RelatedIdentifierType.DOI)],
                descriptions=[CreateIdentifierDescription(description='Test Description')],
                creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            self.assertEqual(exp, response)

    def test_create_identifier_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/identifier', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_identifier(
                    database_id=1, type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except MalformedError:
                pass

    def test_create_identifier_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/identifier', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_identifier(
                    database_id=1, type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except ForbiddenError:
                pass

    def test_create_identifier_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/identifier', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_identifier(
                    database_id=1, type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except NotExistsError:
                pass

    def test_create_identifier_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/identifier', status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.create_identifier(
                    database_id=1, type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except ExternalSystemError:
                pass

    def test_create_identifier_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/identifier', status_code=503)
            # test
            try:
                response = RestClient().create_identifier(
                    database_id=1, type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except AuthenticationError:
                pass

    def test_suggest_identifier_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id=10,
                             database_id=1,
                             publication_year=2024,
                             publisher='TU Wien',
                             titles=[IdentifierTitle(id=10, title='Test Title')],
                             descriptions=[IdentifierDescription(id=10, description='Test')],
                             created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                             last_modified=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                             type=IdentifierType.VIEW,
                             creators=[IdentifierCreator(id=5, creator_name='Carberry, Josiah',
                                                         name_identifier='https://orcid.org/0000-0002-1825-0097')],
                             status=IdentifierStatusType.DRAFT,
                             creator=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')
                             )
            # mock
            mock.get('/api/identifier?url=https://orcid.org/0000-0002-1825-0097', json=exp.model_dump())
            # test
            response = RestClient().suggest_identifier("https://orcid.org/0000-0002-1825-0097")
            self.assertEqual(exp, response)

    def test_suggest_identifier_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/identifier?url=https://orcid.org/0000-0002-1825-0097', status_code=404)
            # test
            try:
                response = RestClient().suggest_identifier("https://orcid.org/0000-0002-1825-0097")
            except NotExistsError:
                pass

    def test_get_identifiers_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [Identifier(id=10,
                              database_id=1,
                              view_id=32,
                              publication_year=2024,
                              publisher='TU Wien',
                              created=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                              last_modified=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                              type=IdentifierType.VIEW,
                              language=Language.EN,
                              descriptions=[IdentifierDescription(id=2, description='Test Description')],
                              titles=[IdentifierTitle(id=3, title='Test Title')],
                              funders=[IdentifierFunder(id=4, funder_name='FWF')],
                              related_identifiers=[RelatedIdentifier(id=7, value='10.12345/abc',
                                                                     relation=RelatedIdentifierRelation.CITES,
                                                                     type=RelatedIdentifierType.DOI)],
                              creators=[IdentifierCreator(id=5, creator_name='Carberry, Josiah')],
                              status=IdentifierStatusType.PUBLISHED,
                              creator=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))]
            # mock
            mock.get('/api/pid', json=[exp[0].model_dump()], headers={"Accept": "application/json"})
            # test
            response = RestClient().get_identifiers()
            self.assertEqual(exp, response)

    def test_get_identifiers_ld_json_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{"@context": "https://schema.org/", "@type": "Dataset", "url": "http://localhost/database/2/info",
                    "citation": "http://localhost/pid/2", "hasPart": [], "version": "2024-03-21T12:05:46.000Z",
                    "name": "sdfsdf", "description": "sfsdf", "identifier": ["http://localhost/pid/2"],
                    "license": "https://creativecommons.org/licenses/by/4.0/legalcode", "creator": [
                    {"name": "Weise, Martin", "@type": "Person", "sameAs": "https://orcid.org/0000-0003-4216-302X",
                     "givenName": "Martin", "familyName": "Weise"}], "temporalCoverage": 2024}]
            # mock
            mock.get('/api/pid', json=exp, headers={"Accept": "application/ld+json"})
            # test
            response = RestClient().get_identifiers(ld=True)
            self.assertEqual(exp, response)


if __name__ == "__main__":
    unittest.main()



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import User, UserAttributes, UserBrief
from dbrepo.api.exceptions import ResponseCodeError, UsernameExistsError, EmailExistsError, NotExistsError, \
    ForbiddenError, AuthenticationError


class UserUnitTest(unittest.TestCase):

    def test_whoami_fails(self):
        username = RestClient().whoami()
        self.assertIsNone(username)

    def test_whoami_succeeds(self):
        client = RestClient(username="a", password="b")
        username = client.whoami()
        self.assertEqual("a", username)

    def test_get_users_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('http://gateway-service/api/user', json=[])
            # test
            response = RestClient().get_users()
            self.assertEqual([], response)

    def test_get_user_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [
                User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                     attributes=UserAttributes(theme='dark'))
            ]
            # mock
            mock.get('http://gateway-service/api/user', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_users()
            self.assertEqual(exp, response)

    def test_get_user_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('http://gateway-service/api/user', status_code=404)
            # test
            try:
                response = RestClient().get_users()
            except ResponseCodeError as e:
                pass

    def test_create_user_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')
            # mock
            mock.post('http://gateway-service/api/user', json=exp.model_dump(), status_code=201)
            # test
            response = RestClient().create_user(username='mweise', password='s3cr3t', email='mweise@example.com')
            self.assertEqual(exp, response)

    def test_create_user_bad_request_fails(self):
        with requests_mock.Mocker() as mock:
            exp = UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')
            # mock
            mock.post('http://gateway-service/api/user', json=exp.model_dump(), status_code=400)
            # test
            try:
                response = RestClient().create_user(username='mweise', password='s3cr3t', email='mweise@example.com')
            except ResponseCodeError as e:
                pass

    def test_create_user_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            exp = UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')
            # mock
            mock.post('http://gateway-service/api/user', json=exp.model_dump(), status_code=403)
            # test
            try:
                response = RestClient().create_user(username='mweise', password='s3cr3t', email='mweise@example.com')
            except ForbiddenError as e:
                pass

    def test_create_user_username_exists_fails(self):
        with requests_mock.Mocker() as mock:
            exp = UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')
            # mock
            mock.post('http://gateway-service/api/user', json=exp.model_dump(), status_code=409)
            # test
            try:
                response = RestClient().create_user(username='mweise', password='s3cr3t', email='mweise@example.com')
            except UsernameExistsError as e:
                pass

    def test_create_user_default_role_not_exists_fails(self):
        with requests_mock.Mocker() as mock:
            exp = UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')
            # mock
            mock.post('http://gateway-service/api/user', json=exp.model_dump(), status_code=404)
            # test
            try:
                response = RestClient().create_user(username='mweise', password='s3cr3t', email='mweise@example.com')
            except NotExistsError as e:
                pass

    def test_create_user_emails_exists_fails(self):
        with requests_mock.Mocker() as mock:
            exp = UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')
            # mock
            mock.post('http://gateway-service/api/user', json=exp.model_dump(), status_code=417)
            # test
            try:
                response = RestClient().create_user(username='mweise', password='s3cr3t', email='mweise@example.com')
            except EmailExistsError as e:
                pass

    def test_get_user_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                       attributes=UserAttributes(theme='dark'))
            # mock
            mock.get('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16',
                     json=exp.model_dump())
            # test
            response = RestClient().get_user(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16')
            self.assertEqual(exp, response)

    def test_get_user_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16', status_code=404)
            # test
            try:
                response = RestClient().get_user(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16')
            except NotExistsError as e:
                pass

    def test_update_user_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise', given_name='Martin',
                       attributes=UserAttributes(theme='dark'))
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16', status_code=202,
                     json=exp.model_dump())
            # test
            client = RestClient(username="a", password="b")
            response = client.update_user(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16', firstname='Martin',
                                          language='en', theme='light')
            self.assertEqual(exp, response)

    def test_update_user_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_user(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16', firstname='Martin',
                                              language='en', theme='light')
            except ForbiddenError as e:
                pass

    def test_update_user_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_user(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16', firstname='Martin',
                                              language='en', theme='light')
            except NotExistsError as e:
                pass

    def test_update_user_foreign_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16', status_code=405)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_user(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16', firstname='Martin',
                                              language='en', theme='light')
            except ForbiddenError as e:
                pass

    def test_update_user_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16', status_code=405)
            # test
            try:
                response = RestClient().update_user(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16', firstname='Martin',
                                                    language='en', theme='light')
            except AuthenticationError as e:
                pass

    def test_update_user_password_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise', given_name='Martin',
                       attributes=UserAttributes(theme='dark'))
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16/password', status_code=202,
                     json=exp.model_dump())
            # test
            client = RestClient(username="a", password="b")
            response = client.update_user_password(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16',
                                                   password='s3cr3t1n0rm4t10n')
            self.assertEqual(exp, response)

    def test_update_user_password_not_allowed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16/password', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_user_password(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16',
                                                       password='s3cr3t1n0rm4t10n')
            except ForbiddenError as e:
                pass

    def test_update_user_password_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16/password', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_user_password(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16',
                                                       password='s3cr3t1n0rm4t10n')
            except NotExistsError as e:
                pass

    def test_update_user_password_foreign_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16/password', status_code=405)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_user_password(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16',
                                                       password='s3cr3t1n0rm4t10n')
            except ForbiddenError as e:
                pass

    def test_update_user_password_keycloak_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16/password', status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                response = client.update_user_password(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16',
                                                       password='s3cr3t1n0rm4t10n')
            except ResponseCodeError as e:
                pass

    def test_update_user_password_not_auth_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('http://gateway-service/api/user/8638c043-5145-4be8-a3e4-4b79991b0a16/password', status_code=503)
            # test
            try:
                response = RestClient().update_user_password(user_id='8638c043-5145-4be8-a3e4-4b79991b0a16',
                                                             password='s3cr3t1n0rm4t10n')
            except AuthenticationError as e:
                pass


if __name__ == "__main__":
    unittest.main()



import sys
import os
import logging

import requests
from pydantic import TypeAdapter
from tusclient.client import TusClient
from pandas import DataFrame

from dbrepo.UploadClient import UploadClient
from dbrepo.api.dto import *
from dbrepo.api.exceptions import ResponseCodeError, UsernameExistsError, EmailExistsError, NotExistsError, \
    ForbiddenError, MalformedError, NameExistsError, QueryStoreError, MetadataConsistencyError, ExternalSystemError, \
    AuthenticationError, UploadError


class RestClient:
    """
    The RestClient class for communicating with the DBRepo REST API. All parameters can be set also via environment \
    variables, e.g. set endpoint with REST_API_ENDPOINT, username with REST_API_USERNAME, etc. You can override \
    the constructor parameters with the environment variables.

    :param endpoint: The REST API endpoint. Optional. Default: "http://gateway-service"
    :param username: The REST API username. Optional.
    :param password: The REST API password. Optional.
    :param secure: When set to false, the requests library will not verify the authenticity of your TLS/SSL
        certificates (i.e. when using self-signed certificates). Default: true.
    """
    endpoint: str = None
    username: str = None
    password: str = None
    secure: bool = None

    def __init__(self,
                 endpoint: str = 'http://gateway-service',
                 username: str = None,
                 password: str = None,
                 secure: bool = True) -> None:
        logging.basicConfig(format='%(asctime)s %(name)-12s %(levelname)-6s %(message)s', level=logging.DEBUG,
                            stream=sys.stdout)
        self.endpoint = os.environ.get('REST_API_ENDPOINT', endpoint)
        self.username = os.environ.get('REST_API_USERNAME', username)
        self.password = os.environ.get('REST_API_PASSWORD', password)
        if os.environ.get('REST_API_SECURE') is not None:
            self.secure = os.environ.get('REST_API_SECURE') == 'True'
        else:
            self.secure = secure

    def _wrapper(self, method: str, url: str, params: [(str,)] = None, payload=None, headers: dict = None,
                 force_auth: bool = False, stream: bool = False) -> requests.Response:
        if force_auth and (self.username is None and self.password is None):
            raise AuthenticationError(f"Failed to perform request: authentication required")
        url = f'{self.endpoint}{url}'
        logging.debug(f'method: {method}')
        logging.debug(f'url: {url}')
        if params is not None:
            logging.debug(f'params: {params}')
        if stream is not None:
            logging.debug(f'stream: {stream}')
        logging.debug(f'secure: {self.secure}')
        if headers is not None:
            logging.debug(f'headers: {headers}')
        else:
            headers = dict()
            logging.debug(f'no headers set')
        if payload is not None:
            payload = payload.model_dump()
        auth = None
        if self.username is None and self.password is not None:
            headers["Authorization"] = f"Bearer {self.password}"
            logging.debug(f'configured for oidc/bearer auth')
        elif self.username is not None and self.password is not None:
            auth = (self.username, self.password)
            logging.debug(f'configured for basic auth: username={self.username}, password=(hidden)')
        return requests.request(method=method, url=url, auth=auth, verify=self.secure,
                                json=payload, headers=headers, params=params, stream=stream)

    def upload(self, file_path: str) -> str:
        """
        Uploads a file located at file_path to the Upload Service.

        :param file_path: The location of the file on the local filesystem.

        :returns: Filename on the S3 backend of the Upload Service, if successful.
        """
        my_client = TusClient(url=f'{self.endpoint}/api/upload/files/')
        uploader = my_client.uploader(file_path=file_path)
        uploader.upload()
        filename = uploader.url[uploader.url.rfind('/') + 1:uploader.url.rfind('+')]
        if filename is None or len(filename) == 0:
            raise UploadError(f'Failed to upload the file to {self.endpoint}')
        return filename

    def get_jwt_auth(self, username: str = None, password: str = None) -> JwtAuth:
        """
        Obtains a JWT auth object from the Auth Service containing e.g. the access token and refresh token.

        :param username: The username used to authenticate with the Auth Service. Optional. Default: username from the `RestClient` constructor.
        :param password: The password used to authenticate with the Auth Service. Optional. Default: password from the `RestClient` constructor.

        :returns: JWT auth object from the Auth Service, if successful.

        :raises ForbiddenError: If something went wrong with the authentication.
        :raises ResponseCodeError: If something went wrong with the authentication.
        """
        if username is None:
            username = self.username
        if password is None:
            password = self.password
        url = f'{self.endpoint}/api/user/token'
        response = requests.post(url=url, json=dict({"username": username, "password": password}))
        if response.status_code == 202:
            body = response.json()
            return JwtAuth.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get JWT auth')
        raise ResponseCodeError(f'Failed to get JWT auth: response code: {response.status_code} is not 202 (ACCEPTED)')

    def whoami(self) -> str | None:
        """
        Print the username.

        :returns: The username, if set.
        """
        if self.username is not None:
            logging.info(f"{self.username}")
            return self.username
        logging.info(f"No username set!")
        return None

    def get_users(self) -> List[UserBrief]:
        """
        Get all users.

        :returns: List of users, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        """
        url = f'/api/user'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return TypeAdapter(List[UserBrief]).validate_python(body)
        raise ResponseCodeError(f'Failed to find users: response code: {response.status_code} is not 200 (OK)')

    def get_user(self, user_id: str) -> User:
        """
        Get a user with given user id.

        :returns: The user, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises NotExistsError: If theuser does not exist.
        """
        url = f'/api/user/{user_id}'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return User.model_validate(body)
        if response.status_code == 404:
            raise NotExistsError(f'Failed to find user with id {user_id}')
        raise ResponseCodeError(
            f'Failed to find user with id {user_id}: response code: {response.status_code} is not 200 (OK)')

    def create_user(self, username: str, password: str, email: str) -> UserBrief:
        """
        Creates a new user.

        :param username: The username of the new user. Must be unique.
        :param password: The password of the new user.
        :param email: The email of the new user. Must be unique.

        :returns: The user, if successful.

        :raises ResponseCodeError: If something went wrong with the creation.
        :raises UsernameExistsError: The username exists already.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedefault role was not found.
        :raises EmailExistsError: The email exists already.
        """
        url = f'/api/user'
        response = self._wrapper(method="post", url=url,
                                 payload=CreateUser(username=username, password=password, email=email))
        if response.status_code == 201:
            body = response.json()
            return UserBrief.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to update user password: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to create user: default role not found')
        if response.status_code == 409:
            raise UsernameExistsError(f'Failed to create user: user with username exists')
        if response.status_code == 417:
            raise EmailExistsError(f'Failed to create user: user with e-mail exists')
        raise ResponseCodeError(
            f'Failed to create user: response code: {response.status_code} is not 201 (CREATED)')

    def update_user(self, user_id: str, theme: str, language: str, firstname: str = None, lastname: str = None,
                    affiliation: str = None, orcid: str = None) -> User:
        """
        Updates a user with given user id.

        :param user_id: The user id of the user that should be updated.
        :param theme: The user theme. One of "light", "dark", "light-contrast", "dark-contrast".
        :param language: The user language localization. One of "en", "de".
        :param firstname: The updated given name. Optional.
        :param lastname: The updated family name. Optional.
        :param affiliation: The updated affiliation identifier. Optional.
        :param orcid: The updated ORCID identifier. Optional.

        :returns: The user, if successful.

        :raises ResponseCodeError: If something went wrong with the update.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If theuser does not exist.
        """
        url = f'/api/user/{user_id}'
        response = self._wrapper(method="put", url=url, force_auth=True,
                                 payload=UpdateUser(theme=theme, language=language, firstname=firstname,
                                                    lastname=lastname, affiliation=affiliation, orcid=orcid))
        if response.status_code == 202:
            body = response.json()
            return User.model_validate(body)
        if response.status_code == 400:
            raise ResponseCodeError(f'Failed to update user: invalid values')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to update user password: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update user: user not found')
        if response.status_code == 405:
            raise ForbiddenError(f'Failed to update user: foreign user')
        raise ResponseCodeError(
            f'Failed to update user: response code: {response.status_code} is not 202 (ACCEPTED)')

    def update_user_password(self, user_id: str, password: str) -> User:
        """
        Updates the password of a user with given user id.

        :param user_id: The user id of the user that should be updated.
        :param password: The updated user password.

        :returns: The user, if successful.

        :raises ResponseCodeError: If something went wrong with the update.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If theuser does not exist.
        """
        url = f'/api/user/{user_id}/password'
        response = self._wrapper(method="put", url=url, force_auth=True, payload=UpdateUserPassword(password=password))
        if response.status_code == 202:
            body = response.json()
            return User.model_validate(body)
        if response.status_code == 400:
            raise ResponseCodeError(f'Failed to update user password: invalid values')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to update user password: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update user password: not found')
        if response.status_code == 405:
            raise ForbiddenError(f'Failed to update user password: foreign user')
        if response.status_code == 503:
            raise ResponseCodeError(f'Failed to update user password: keycloak error')
        raise ResponseCodeError(
            f'Failed to update user theme: response code: {response.status_code} is not 202 (ACCEPTED)')

    def get_containers(self) -> List[ContainerBrief]:
        """
        Get all containers.

        :returns: List of containers, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        """
        url = f'/api/container'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return TypeAdapter(List[ContainerBrief]).validate_python(body)
        raise ResponseCodeError(f'Failed to find containers: response code: {response.status_code} is not 200 (OK)')

    def get_container(self, container_id: int) -> Container:
        """
        Get a container with given id.

        :returns: List of containers, if successful.

        :raises NotExistsError: If the container does not exist.
        :raises ResponseCodeError: If something went wrong with the retrieval.
        """
        url = f'/api/container/{container_id}'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return Container.model_validate(body)
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get container: not found')
        raise ResponseCodeError(f'Failed to get container: response code: {response.status_code} is not 200 (OK)')

    def get_databases(self) -> List[Database]:
        """
        Get all databases.

        :returns: List of databases, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        """
        url = f'/api/database'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return TypeAdapter(List[Database]).validate_python(body)
        raise ResponseCodeError(f'Failed to find databases: response code: {response.status_code} is not 200 (OK)')

    def get_databases_count(self) -> int:
        """
        Count all databases.

        :returns: Count of databases if successful.
        :raises ResponseCodeError: If something went wrong with the retrieval.
        """
        url = f'/api/database'
        response = self._wrapper(method="head", url=url)
        if response.status_code == 200:
            return int(response.headers.get("x-count"))
        raise ResponseCodeError(f'Failed to find databases: response code: {response.status_code} is not 200 (OK)')

    def get_database(self, database_id: int) -> Database:
        """
        Get a databases with given id.

        :param database_id: The database id.

        :returns: The database, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        """
        url = f'/api/database/{database_id}'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return Database.model_validate(body)
        if response.status_code == 404:
            raise NotExistsError(f'Failed to find database with id {database_id}')
        raise ResponseCodeError(
            f'Failed to find database with id {database_id}: response code: {response.status_code} is not 200 (OK)')

    def create_database(self, name: str, container_id: int, is_public: bool) -> Database:
        """
        Create a databases in a container with given container id.

        :param name: The name of the database.
        :param container_id: The container id.
        :param is_public: The visibility of the database. If set to true everything will be visible, otherwise only
                the metadata (schema, identifiers) will be visible to the public.

        :returns: The database, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the container does not exist.
        """
        url = f'/api/database'
        response = self._wrapper(method="post", url=url, force_auth=True,
                                 payload=CreateDatabase(name=name, container_id=container_id, is_public=is_public))
        if response.status_code == 201:
            body = response.json()
            return Database.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to create database: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to create database: container not found')
        raise ResponseCodeError(
            f'Failed to create database: response code: {response.status_code} is not 201 (CREATED)')

    def update_database_visibility(self, database_id: int, is_public: bool) -> Database:
        """
        Updates the database visibility of a database with given database id.

        :param database_id: The database id.
        :param is_public: The visibility of the database. If set to true everything will be visible, otherwise only
                the metadata (schema, identifiers) will be visible to the public.

        :returns: The database, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedatabase does not exist.
        """
        url = f'/api/database/{database_id}'
        response = self._wrapper(method="put", url=url, force_auth=True, payload=ModifyVisibility(is_public=is_public))
        if response.status_code == 202:
            body = response.json()
            return Database.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to update database visibility: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update database visibility: not found')
        raise ResponseCodeError(
            f'Failed to update database visibility: response code: {response.status_code} is not 202 (ACCEPTED)')

    def update_database_owner(self, database_id: int, user_id: str) -> Database:
        """
        Updates the database owner of a database with given database id.

        :param database_id: The database id.
        :param user_id: The user id of the new owner.

        :returns: The database, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises NotExistsError: If thedatabase does not exist.
        """
        url = f'/api/database/{database_id}/owner'
        response = self._wrapper(method="put", url=url, force_auth=True, payload=ModifyOwner(id=user_id))
        if response.status_code == 202:
            body = response.json()
            return Database.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to update database visibility: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update database visibility: not found')
        raise ResponseCodeError(
            f'Failed to update database visibility: response code: {response.status_code} is not 202 (ACCEPTED)')

    def create_table(self, database_id: int, name: str, columns: List[CreateTableColumn],
                     constraints: CreateTableConstraints, description: str = None) -> Table:
        """
        Updates the database owner of a database with given database id.

        :param database_id: The database id.
        :param name: The name of the created table.
        :param constraints: The constraints of the created table.
        :param columns: The columns of the created table.
        :param description: The description of the created table. Optional.

        :returns: The table, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises NameExistsError: If a table with this name already exists.
        :raises ForbiddenError: If the action is not allowed.
        :raises MalformedError: If the payload is rejected by the service.
        :raises NotExistsError: If the container does not exist.
        """
        url = f'/api/database/{database_id}/table'
        response = self._wrapper(method="post", url=url, force_auth=True,
                                 payload=CreateTable(name=name, description=description,
                                                     columns=columns, constraints=constraints))
        if response.status_code == 201:
            body = response.json()
            return Table.model_validate(body)
        if response.status_code == 400:
            raise MalformedError(f'Failed to create table: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to create table: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to create table: container not found')
        if response.status_code == 409:
            raise NameExistsError(f'Failed to create table: table name exists')
        raise ResponseCodeError(
            f'Failed to create table: response code: {response.status_code} is not 201 (CREATED)')

    def get_tables(self, database_id: int) -> List[TableBrief]:
        """
        Get all tables.

        :param database_id: The database id.

        :returns: List of tables, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        """
        url = f'/api/database/{database_id}/table'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return TypeAdapter(List[TableBrief]).validate_python(body)
        raise ResponseCodeError(f'Failed to find tables: response code: {response.status_code} is not 200 (OK)')

    def get_table(self, database_id: int, table_id: int) -> Table:
        """
        Get a table with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.

        :returns: List of tables, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/table/{table_id}'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return Table.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to find table: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to find table: not found')
        raise ResponseCodeError(f'Failed to find table: response code: {response.status_code} is not 200 (OK)')

    def delete_table(self, database_id: int, table_id: int) -> None:
        """
        Delete a table with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/table/{table_id}'
        response = self._wrapper(method="delete", url=url, force_auth=True)
        if response.status_code == 202:
            return
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to delete table: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to delete table: not found')
        raise ResponseCodeError(f'Failed to delete table: response code: {response.status_code} is not 202 (ACCEPTED)')

    def get_table_metadata(self, database_id: int) -> Database:
        """
        Generate metadata of all system-versioned tables in a database with given id.

        :param database_id: The database id.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the container does not exist.
        """
        url = f'/api/database/{database_id}/metadata/table'
        response = self._wrapper(method="put", url=url, force_auth=True)
        if response.status_code == 200:
            body = response.json()
            return Database.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get tables metadata: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get tables metadata: not found')
        raise ResponseCodeError(f'Failed to get tables metadata: response code: {response.status_code} is not 200 (OK)')

    def get_views(self, database_id: int) -> List[View]:
        """
        Gets views of a database with given database id.

        :param database_id: The database id.

        :returns: The list of views, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/view'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return TypeAdapter(List[View]).validate_python(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to find views: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to find views: not found')
        raise ResponseCodeError(f'Failed to find views: response code: {response.status_code} is not 200 (OK)')

    def get_view(self, database_id: int, view_id: int) -> View:
        """
        Get a view of a database with given database id and view id.

        :param database_id: The database id.
        :param view_id: The view id.

        :returns: The view, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/view/{view_id}'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return View.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to find view: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to find view: not found')
        raise ResponseCodeError(f'Failed to find view: response code: {response.status_code} is not 200 (OK)')

    def create_view(self, database_id: int, name: str, query: str, is_public: bool) -> View:
        """
        Create a view in a database with given database id.

        :param database_id: The database id.
        :param name: The name of the created view.
        :param query: The query of the created view.
        :param is_public: The visibility of the view. If set to true everything will be visible, otherwise only
                the metadata (schema, identifiers) will be visible to the public.

        :returns: The created view, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/view'
        response = self._wrapper(method="post", url=url, force_auth=True,
                                 payload=CreateView(name=name, query=query, is_public=is_public))
        if response.status_code == 201:
            body = response.json()
            return View.model_validate(body)
        if response.status_code == 400 or response.status_code == 423:
            raise MalformedError(f'Failed to create view: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to create view: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to create view: not found')
        raise ResponseCodeError(f'Failed to create view: response code: {response.status_code} is not 201 (CREATED)')

    def delete_view(self, database_id: int, view_id: int) -> None:
        """
        Deletes a view in a database with given database id and view id.

        :param database_id: The database id.
        :param view_id: The view id.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/view/{view_id}'
        response = self._wrapper(method="delete", url=url, force_auth=True)
        if response.status_code == 202:
            return
        if response.status_code == 400 or response.status_code == 423:
            raise MalformedError(f'Failed to delete view: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to delete view: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to delete view: not found')
        raise ResponseCodeError(f'Failed to delete view: response code: {response.status_code} is not 202 (ACCEPTED)')

    def get_view_data(self, database_id: int, view_id: int, page: int = 0, size: int = 10,
                      df: bool = False) -> Result | DataFrame:
        """
        Get data of a view in a database with given database id and view id.

        :param database_id: The database id.
        :param view_id: The view id.
        :param page: The result pagination number. Optional. Default: 0.
        :param size: The result pagination size. Optional. Default: 10.
        :param df: If true, the result is returned as Pandas DataFrame. Optional. Default: False.

        :returns: The result of the view query, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the view does not exist.
        """
        url = f'/api/database/{database_id}/view/{view_id}/data'
        params = []
        if page is not None and size is not None:
            params.append(('page', page))
            params.append(('size', size))
        response = self._wrapper(method="get", url=url, params=params)
        if response.status_code == 200:
            body = response.json()
            res = Result.model_validate(body)
            if df:
                return DataFrame.from_records(res.result)
            return res
        if response.status_code == 400:
            raise MalformedError(f'Failed to get view data: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get view data: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get view data: not found')
        raise ResponseCodeError(f'Failed to get view data: response code: {response.status_code} is not 200 (OK)')

    def get_views_metadata(self, database_id: int) -> Database:
        """
        Generate metadata of all views in a database with given id.

        :param database_id: The database id.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the container does not exist.
        """
        url = f'/api/database/{database_id}/metadata/view'
        response = self._wrapper(method="put", url=url, force_auth=True)
        if response.status_code == 200:
            body = response.json()
            return Database.model_validate(body)
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get views metadata: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get views metadata: not found')
        raise ResponseCodeError(f'Failed to get views metadata: response code: {response.status_code} is not 200 (OK)')

    def get_table_data(self, database_id: int, table_id: int, page: int = 0, size: int = 10,
                       timestamp: datetime.datetime = None, df: bool = False) -> Result | DataFrame:
        """
        Get data of a table in a database with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.
        :param page: The result pagination number. Optional. Default: 0.
        :param size: The result pagination size. Optional. Default: 10.
        :param timestamp: The query execution time. Optional.
        :param df: If true, the result is returned as Pandas DataFrame. Optional. Default: False.

        :returns: The result of the view query, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the table does not exist.
        :raises QueryStoreError: If the result set could not be counted.
        """
        url = f'/api/database/{database_id}/table/{table_id}/data'
        params = []
        if page is not None and size is not None:
            params.append(('page', page))
            params.append(('size', size))
        if timestamp is not None:
            params.append(('timestamp', timestamp))
        response = self._wrapper(method="get", url=url, params=params)
        if response.status_code == 200:
            body = response.json()
            res = Result.model_validate(body)
            if df:
                return DataFrame.from_records(res.result)
            return res
        if response.status_code == 400:
            raise MalformedError(f'Failed to get table data: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get table data: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get table data: not found')
        if response.status_code == 409:
            raise QueryStoreError(f'Failed to get table data: service rejected result count')
        raise ResponseCodeError(f'Failed to get table data: response code: {response.status_code} is not 200 (OK)')

    def create_table_data(self, database_id: int, table_id: int, data: dict) -> None:
        """
        Insert data into a table in a database with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.
        :param data: The data dictionary to be inserted into the table with the form column=value of the table.

        :raises ResponseCodeError: If something went wrong with the insert.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the table does not exist.
        :raises MalformedError: If the payload is rejected by the service (e.g. LOB data could not be imported).
        """
        url = f'/api/database/{database_id}/table/{table_id}/data'
        response = self._wrapper(method="post", url=url, force_auth=True, payload=CreateData(data=data))
        if response.status_code == 201:
            return
        if response.status_code == 400 or response.status_code == 410:
            raise MalformedError(f'Failed to insert table data: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to insert table data: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to insert table data: not found')
        raise ResponseCodeError(
            f'Failed to insert table data: response code: {response.status_code} is not 201 (CREATED)')

    def import_table_data(self, database_id: int, table_id: int, separator: str, file_path: str,
                          quote: str = None, skip_lines: int = 0, false_encoding: str = None,
                          true_encoding: str = None, null_encoding: str = None,
                          line_encoding: str = "\r\n") -> None:
        """
        Import a csv dataset from a file into a table in a database with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.
        :param separator: The csv column separator.
        :param file_path: The path of the file that is imported on the storage service.
        :param quote: The column data quotation character. Optional.
        :param skip_lines: The number of lines to skip. Optional. Default: 0.
        :param false_encoding: The encoding of boolean false. Optional.
        :param true_encoding: The encoding of boolean true. Optional.
        :param null_encoding: The encoding of null. Optional.
        :param line_encoding: The encoding of the line termination. Optional. Default: CR (Windows).

        :raises ResponseCodeError: If something went wrong with the insert.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the table does not exist.
        :raises MalformedError: If the payload is rejected by the service (e.g. LOB data could not be imported).
        """
        client = UploadClient(endpoint=f"{self.endpoint}/api/upload/files")
        filename = client.upload(file_path=file_path)
        url = f'/api/database/{database_id}/table/{table_id}/data/import'
        response = self._wrapper(method="post", url=url, force_auth=True,
                                 payload=Import(location=filename, separator=separator, quote=quote,
                                                skip_lines=skip_lines, false_element=false_encoding,
                                                true_element=true_encoding, null_element=null_encoding,
                                                line_termination=line_encoding))
        if response.status_code == 202:
            return
        if response.status_code == 400:
            raise MalformedError(f'Failed to import table data: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to import table data: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to import table data: not found')
        if response.status_code == 409 or response.status_code == 422:
            raise ExternalSystemError(f'Failed to import table data: sidecar rejected the import')
        raise ResponseCodeError(
            f'Failed to import table data: response code: {response.status_code} is not 202 (ACCEPTED)')

    def analyse_datatypes(self, file_path: str, separator: str, enum: bool = None,
                          enum_tol: int = None, upload: bool = True) -> DatatypeAnalysis:
        """
        Import a csv dataset from a file and analyse it for the possible enums, line encoding and column data types.

        :param file_path: The path of the file that is imported on the storage service.
        :param separator: The csv column separator.
        :param enum: If set to true, enumerations should be guessed, otherwise no guessing. Optional.
        :param enum_tol: The tolerance for guessing enumerations (ignored if enum=False). Optional.
        :param upload: If set to true, the file from file_path will be uploaded, otherwise no upload will be performed \
            and the file_path will be treated as S3 filename and analysed instead. Optional. Default: true.

        :returns: The determined data types, if successful.

        :raises ResponseCodeError: If something went wrong with the analysis.
        :raises MalformedError: If the payload is rejected by the service.
        :raises NotExistsError: If the file was not found by the Analyse Service.
        """
        if upload:
            client = UploadClient(endpoint=f"{self.endpoint}/api/upload/files")
            filename = client.upload(file_path=file_path)
        else:
            filename = file_path
        params = [
            ('filename', filename),
            ('separator', separator),
            ('enum', enum),
            ('enum_tol', enum_tol)
        ]
        url = f'/api/analyse/datatypes'
        response = self._wrapper(method="get", url=url, params=params)
        if response.status_code == 202:
            body = response.json()
            return DatatypeAnalysis.model_validate(body)
        if response.status_code == 400 or response.status_code == 500:
            raise MalformedError(f'Failed to analyse data types: service rejected malformed payload')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to analyse data types: failed to find file in Storage Service')
        raise ResponseCodeError(
            f'Failed to analyse data types: response code: {response.status_code} is not 202 (ACCEPTED)')

    def analyse_keys(self, file_path: str, separator: str, upload: bool = True) -> KeyAnalysis:
        """
        Import a csv dataset from a file and analyse it for the possible primary key.

        :param file_path: The path of the file that is imported on the storage service.
        :param separator: The csv column separator.
        :param upload: If set to true, the file from file_path will be uploaded, otherwise no upload will be performed \
            and the file_path will be treated as S3 filename and analysed instead. Optional. Default: true.

        :returns: The determined ranking of the primary key candidates, if successful.

        :raises ResponseCodeError: If something went wrong with the analysis.
        :raises MalformedError: If the payload is rejected by the service.
        :raises NotExistsError: If the file was not found by the Analyse Service.
        """
        if upload:
            client = UploadClient(endpoint=f"{self.endpoint}/api/upload/files")
            filename = client.upload(file_path=file_path)
        else:
            filename = file_path
        params = [
            ('filename', filename),
            ('separator', separator),
        ]
        url = f'/api/analyse/keys'
        response = self._wrapper(method="get", url=url, params=params)
        if response.status_code == 202:
            body = response.json()
            return KeyAnalysis.model_validate(body)
        if response.status_code == 400 or response.status_code == 500:
            raise MalformedError(f'Failed to analyse data types: service rejected malformed payload')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to analyse data types: failed to find file in Storage Service')
        raise ResponseCodeError(
            f'Failed to analyse data types: response code: {response.status_code} is not 202 (ACCEPTED)')

    def analyse_table_statistics(self, database_id: int, table_id: int) -> TableStatistics:
        """
        Analyses the numerical contents of a table in a database with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.

        :returns: The table statistics, if successful.

        :raises ResponseCodeError: If something went wrong with the analysis.
        :raises MalformedError: If the payload is rejected by the service.
        :raises NotExistsError: If the file was not found by the Analyse Service.
        """
        url = f'/api/analyse/database/{database_id}/table/{table_id}/statistics'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 202:
            body = response.json()
            return TableStatistics.model_validate(body)
        if response.status_code == 400:
            raise MalformedError(f'Failed to analyse table statistics: service rejected malformed payload')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to analyse table statistics: separator error')
        raise ResponseCodeError(
            f'Failed to analyse table statistics: response code: {response.status_code} is not 202 (ACCEPTED)')

    def update_table_data(self, database_id: int, table_id: int, data: dict, keys: dict) -> None:
        """
        Update data in a table in a database with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.
        :param data: The data dictionary to be updated into the table with the form column=value of the table.
        :param keys: The key dictionary matching the rows in the form column=value.

        :raises ResponseCodeError: If something went wrong with the update.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the table does not exist.
        :raises MalformedError: If the payload is rejected by the service (e.g. LOB data could not be imported).
        """
        url = f'/api/database/{database_id}/table/{table_id}/data'
        response = self._wrapper(method="put", url=url, force_auth=True, payload=UpdateData(data=data, keys=keys))
        if response.status_code == 202:
            return
        if response.status_code == 400 or response.status_code == 410:
            raise MalformedError(f'Failed to update table data: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to update table data: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update table data: not found')
        raise ResponseCodeError(
            f'Failed to update table data: response code: {response.status_code} is not 202 (ACCEPTED)')

    def delete_table_data(self, database_id: int, table_id: int, keys: dict) -> None:
        """
        Delete data in a table in a database with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.
        :param keys: The key dictionary matching the rows in the form column=value.

        :raises ResponseCodeError: If something went wrong with the deletion.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the table does not exist.
        :raises MalformedError: If the payload is rejected by the service.
        """
        url = f'/api/database/{database_id}/table/{table_id}/data'
        response = self._wrapper(method="delete", url=url, force_auth=True, payload=DeleteData(keys=keys))
        if response.status_code == 202:
            return
        if response.status_code == 400:
            raise MalformedError(f'Failed to delete table data: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to delete table data: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to delete table data: not found')
        raise ResponseCodeError(
            f'Failed to delete table data: response code: {response.status_code} is not 202 (ACCEPTED)')

    def get_table_data_count(self, database_id: int, table_id: int, page: int = 0, size: int = 10,
                             timestamp: datetime.datetime = None) -> int:
        """
        Get data count of a table in a database with given database id and table id.

        :param database_id: The database id.
        :param table_id: The table id.
        :param page: The result pagination number. Optional. Default: 0.
        :param size: The result pagination size. Optional. Default: 10.
        :param timestamp: The query execution time. Optional.

        :returns: The result of the view query, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the table does not exist.
        :raises QueryStoreError: If the result set could not be counted.
        """
        url = f'/api/database/{database_id}/table/{table_id}/data'
        if page is not None and size is not None:
            url += f'?page={page}&size={size}'
        if timestamp is not None:
            if page is not None and size is not None:
                url += '&'
            else:
                url += '?'
            url += f'timestamp={timestamp}'
        response = self._wrapper(method="head", url=url)
        if response.status_code == 200:
            return int(response.headers.get('X-Count'))
        if response.status_code == 400:
            raise MalformedError(f'Failed to get table data: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get table data: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get table data: not found')
        if response.status_code == 409:
            raise QueryStoreError(f'Failed to get table data: service rejected result count')
        raise ResponseCodeError(f'Failed to get table data: response code: {response.status_code} is not 200 (OK)')

    def get_view_data_count(self, database_id: int, view_id: int) -> int:
        """
        Get data count of a view in a database with given database id and view id.

        :param database_id: The database id.
        :param view_id: The view id.

        :returns: The result count of the view query, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the payload is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/view/{view_id}/data'
        response = self._wrapper(method="head", url=url)
        if response.status_code == 200:
            return int(response.headers.get('X-Count'))
        if response.status_code == 400:
            raise MalformedError(f'Failed to get view data count: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get view data count: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get view data count: not found')
        raise ResponseCodeError(f'Failed to get view data count: response code: {response.status_code} is not 200 (OK)')

    def get_database_access(self, database_id: int) -> AccessType:
        """
        Get access of a view in a database with given database id and view id.

        :param database_id: The database id.

        :returns: The access type, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thecontainer does not exist.
        """
        url = f'/api/database/{database_id}/access'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return DatabaseAccess.model_validate(body).type
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to get database access: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to get database access: not found')
        raise ResponseCodeError(f'Failed to get database access: response code: {response.status_code} is not 200 (OK)')

    def create_database_access(self, database_id: int, user_id: str, type: AccessType) -> AccessType:
        """
        Create access to a database with given database id and user id.

        :param database_id: The database id.
        :param user_id: The user id.
        :param type: The access type.

        :returns: The access type, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the payload is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedatabase or user does not exist.
        """
        url = f'/api/database/{database_id}/access/{user_id}'
        response = self._wrapper(method="post", url=url, force_auth=True, payload=CreateAccess(type=type))
        if response.status_code == 202:
            body = response.json()
            return DatabaseAccess.model_validate(body).type
        if response.status_code == 400:
            raise MalformedError(f'Failed to create database access: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to create database access: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to create database access: not found')
        raise ResponseCodeError(
            f'Failed to create database access: response code: {response.status_code} is not 202 (ACCEPTED)')

    def update_database_access(self, database_id: int, user_id: str, type: AccessType) -> AccessType:
        """
        Updates the access for a user to a database with given database id and user id.

        :param database_id: The database id.
        :param user_id: The user id.
        :param type: The access type.

        :returns: The access type, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the payload is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedatabase or user does not exist.
        """
        url = f'/api/database/{database_id}/access/{user_id}'
        response = self._wrapper(method="put", url=url, force_auth=True, payload=UpdateAccess(type=type))
        if response.status_code == 202:
            body = response.json()
            return DatabaseAccess.model_validate(body).type
        if response.status_code == 400:
            raise MalformedError(f'Failed to update database access: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to update database access: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update database access: not found')
        raise ResponseCodeError(
            f'Failed to update database access: response code: {response.status_code} is not 202 (ACCEPTED)')

    def delete_database_access(self, database_id: int, user_id: str) -> None:
        """
        Deletes the access for a user to a database with given database id and user id.

        :param database_id: The database id.
        :param user_id: The user id.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the payload is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedatabase or user does not exist.
        """
        url = f'/api/database/{database_id}/access/{user_id}'
        response = self._wrapper(method="delete", url=url, force_auth=True)
        if response.status_code == 202:
            return
        if response.status_code == 400:
            raise MalformedError(f'Failed to delete database access: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to delete database access: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to delete database access: not found')
        raise ResponseCodeError(
            f'Failed to delete database access: response code: {response.status_code} is not 202 (ACCEPTED)')

    def execute_query(self, database_id: int, query: str, page: int = 0, size: int = 10,
                      timestamp: datetime.datetime = datetime.datetime.now()) -> Result:
        """
        Executes a SQL query in a database where the current user has at least read access with given database id. The
        result set can be paginated with setting page and size (both). Historic data can be queried by setting
        timestamp.

        :param database_id: The database id.
        :param query: The query statement.
        :param page: The result pagination number. Optional. Default: 0.
        :param size: The result pagination size. Optional. Default: 10.
        :param timestamp: The query execution time. Optional.

        :returns: The result set, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the payload is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the database, table or user does not exist.
        :raises QueryStoreError: The query store rejected the query.
        :raises MetadataConsistencyError: The service failed to parse columns from the metadata database.
        """
        url = f'/api/database/{database_id}/subset'
        if page is not None and size is not None:
            url += f'?page={page}&size={size}'
        response = self._wrapper(method="post", url=url, force_auth=True,
                                 payload=ExecuteQuery(statement=query, timestamp=timestamp))
        if response.status_code == 201:
            body = response.json()
            return Result.model_validate(body)
        if response.status_code == 400:
            raise MalformedError(f'Failed to execute query: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to execute query: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to execute query: not found')
        if response.status_code == 409:
            raise QueryStoreError(f'Failed to execute query: query store rejected query')
        if response.status_code == 417:
            raise MetadataConsistencyError(f'Failed to execute query: service expected other metadata')
        raise ResponseCodeError(
            f'Failed to execute query: response code: {response.status_code} is not 202 (ACCEPTED)')

    def get_query_data(self, database_id: int, query_id: int, page: int = 0, size: int = 10,
                       df: bool = False) -> Result | DataFrame:
        """
        Re-executes a query in a database with given database id and query id.

        :param database_id: The database id.
        :param query_id: The query id.
        :param page: The result pagination number. Optional. Default: 0.
        :param size: The result pagination size. Optional. Default: 10.
        :param size: The result pagination size. Optional. Default: 10.
        :param df: If true, the result is returned as Pandas DataFrame. Optional. Default: False.

        :returns: The result set, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the payload is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the database, query or user does not exist.
        :raises QueryStoreError: The query store rejected the query.
        :raises MetadataConsistencyError: The service failed to parse columns from the metadata database.
        """
        headers = {}
        url = f'/api/database/{database_id}/subset/{query_id}/data'
        if page is not None and size is not None:
            url += f'?page={page}&size={size}'
        response = self._wrapper(method="get", url=url, headers=headers)
        if response.status_code == 200:
            body = response.json()
            res = Result.model_validate(body)
            if df:
                return DataFrame.from_records(res.result)
            return res
        if response.status_code == 400:
            raise MalformedError(f'Failed to re-execute query: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to re-execute query: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to re-execute query: not found')
        if response.status_code == 409:
            raise QueryStoreError(f'Failed to re-execute query: query store rejected query')
        if response.status_code == 417:
            raise MetadataConsistencyError(f'Failed to re-execute query: service expected other metadata')
        raise ResponseCodeError(
            f'Failed to re-execute query: response code: {response.status_code} is not 200 (OK)')

    def get_query_data_count(self, database_id: int, query_id: int, page: int = 0, size: int = 10) -> int:
        """
        Re-executes a query in a database with given database id and query id and only counts the results.

        :param database_id: The database id.
        :param query_id: The query id.
        :param page: The result pagination number. Optional. Default: 0.
        :param size: The result pagination size. Optional. Default: 10.

        :returns: The result set, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the payload is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If the database, query or user does not exist.
        :raises QueryStoreError: The query store rejected the query.
        :raises MetadataConsistencyError: The service failed to parse columns from the metadata database.
        """
        url = f'/api/database/{database_id}/subset/{query_id}/data'
        if page is not None and size is not None:
            url += f'?page={page}&size={size}'
        response = self._wrapper(method="head", url=url)
        if response.status_code == 200:
            return int(response.headers.get('X-Count'))
        if response.status_code == 400:
            raise MalformedError(f'Failed to re-execute query: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to re-execute query: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to re-execute query: not found')
        if response.status_code == 409:
            raise QueryStoreError(f'Failed to re-execute query: query store rejected query')
        if response.status_code == 417:
            raise MetadataConsistencyError(f'Failed to re-execute query: service expected other metadata')
        raise ResponseCodeError(
            f'Failed to re-execute query: response code: {response.status_code} is not 200 (OK)')

    def get_query(self, database_id: int, query_id: int) -> Query:
        """
        Get query from a database with given database id and query id.

        :param database_id: The database id.
        :param query_id: The query id.

        :returns: The query, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedatabase, query or user does not exist.
        :raises QueryStoreError: The query store rejected the query.
        :raises MetadataConsistencyError: The service failed to parse columns from the metadata database.
        """
        url = f'/api/database/{database_id}/subset/{query_id}'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return Query.model_validate(body)
        if response.status_code == 404:
            raise NotExistsError(f'Failed to find query: not found')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to find query: not allowed')
        if response.status_code == 417:
            raise MetadataConsistencyError(f'Failed to find query: service expected other metadata')
        if response.status_code == 501 or response.status_code == 503 or response.status_code == 504:
            raise QueryStoreError(f'Failed to find query: query store rejected query')
        raise ResponseCodeError(
            f'Failed to find query: response code: {response.status_code} is not 200 (OK)')

    def get_queries(self, database_id: int) -> List[Query]:
        """
        Get queries from a database with given database id.

        :param database_id: The database id.

        :returns: List of queries, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises MalformedError: If the query is rejected by the service.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedatabase or user does not exist.
        :raises QueryStoreError: The query store rejected the query.
        """
        url = f'/api/database/{database_id}/subset'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return TypeAdapter(List[Query]).validate_python(body)
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to find queries: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to find queries: not found')
        if response.status_code == 423:
            raise MalformedError(f'Failed to find queries: service rejected malformed query')
        if response.status_code == 501 or response.status_code == 503 or response.status_code == 504:
            raise QueryStoreError(f'Failed to find queries: query store rejected query')
        raise ResponseCodeError(
            f'Failed to find query: response code: {response.status_code} is not 200 (OK)')

    def update_query(self, database_id: int, query_id: int, persist: bool) -> Query:
        """
        Update query from a database with given database id and query id.

        :param database_id: The database id.
        :param query_id: The query id.
        :param persist: If set to true, the query will be saved and visible in the User Interface, otherwise the query \
                is marked for deletion in the future and not visible in the User Interface.

        :returns: The query, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval.
        :raises ForbiddenError: If the action is not allowed.
        :raises NotExistsError: If thedatabase or user does not exist.
        :raises QueryStoreError: The query store rejected the update.
        """
        url = f'/api/database/{database_id}/subset/{query_id}'
        response = self._wrapper(method="put", url=url, force_auth=True, payload=UpdateQuery(persist=persist))
        if response.status_code == 202:
            body = response.json()
            return Query.model_validate(body)
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to update query: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update query: not found')
        if response.status_code == 412:
            raise QueryStoreError(f'Failed to update query: query store rejected update')
        raise ResponseCodeError(
            f'Failed to update query: response code: {response.status_code} is not 200 (OK)')

    def create_identifier(self, database_id: int, type: IdentifierType, titles: List[CreateIdentifierTitle],
                          publisher: str, creators: List[CreateIdentifierCreator], publication_year: int,
                          descriptions: List[CreateIdentifierDescription] = None,
                          funders: List[CreateIdentifierFunder] = None, licenses: List[License] = None,
                          language: Language = None, query_id: int = None, view_id: int = None, table_id: int = None,
                          publication_day: int = None, publication_month: int = None,
                          related_identifiers: List[CreateRelatedIdentifier] = None) -> Identifier:
        """
        Create an identifier draft.

        :param database_id: The database id of the created identifier.
        :param type: The type of the created identifier.
        :param titles: The titles of the created identifier.
        :param publisher: The publisher of the created identifier.
        :param creators: The creator(s) of the created identifier.
        :param publication_year: The publication year of the created identifier.
        :param descriptions: The description(s) of the created identifier. Optional.
        :param funders: The funders(s) of the created identifier. Optional.
        :param licenses: The license(s) of the created identifier. Optional.
        :param language: The language of the created identifier. Optional.
        :param query_id: The query id of the created identifier. Required when type=SUBSET, otherwise invalid. Optional.
        :param view_id: The view id of the created identifier. Required when type=VIEW, otherwise invalid. Optional.
        :param table_id: The table id of the created identifier. Required when type=TABLE, otherwise invalid. Optional.
        :param publication_day: The publication day of the created identifier. Optional.
        :param publication_month: The publication month of the created identifier. Optional.
        :param related_identifiers: The related identifier(s) of the created identifier. Optional.

        :returns: The identifier, if successful.

        :raises ResponseCodeError: If something went wrong with the creation of the identifier.
        :raises ForbiddenError: If the action is not allowed.
        :raises MalformedError: If the payload is rejected by the service.
        :raises NotExistsError: If the database, table/view/subset or user does not exist.
        :raises ExternalSystemError: If the external system (DataCite) refused communication with the service.
        """
        url = f'/api/identifier'
        payload = CreateIdentifier(database_id=database_id, type=type, titles=titles, publisher=publisher,
                                   creators=creators, publication_year=publication_year, descriptions=descriptions,
                                   funders=funders, licenses=licenses, language=language, query_id=query_id,
                                   view_id=view_id, table_id=table_id, publication_day=publication_day,
                                   publication_month=publication_month, related_identifiers=related_identifiers)
        response = self._wrapper(method="post", url=url, force_auth=True, payload=payload)
        if response.status_code == 201:
            body = response.json()
            return Identifier.model_validate(body)
        if response.status_code == 400:
            raise MalformedError(f'Failed to create identifier: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to create identifier: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to create identifier: not found')
        if response.status_code == 503:
            raise ExternalSystemError(f'Failed to create identifier: external system rejected communication')
        raise ResponseCodeError(
            f'Failed to create identifier: response code: {response.status_code} is not 201 (CREATED)')

    def save_identifier(self, identifier_id: int, database_id: int, type: IdentifierType,
                        titles: List[CreateIdentifierTitle], publisher: str, creators: List[CreateIdentifierCreator],
                        publication_year: int, descriptions: List[CreateIdentifierDescription] = None,
                        funders: List[CreateIdentifierFunder] = None, licenses: List[License] = None,
                        language: Language = None, query_id: int = None, view_id: int = None, table_id: int = None,
                        publication_day: int = None, publication_month: int = None,
                        related_identifiers: List[CreateRelatedIdentifier] = None) -> Identifier:
        """
        Save an existing identifier and update the metadata attached to it.

        :param identifier_id: The identifier id.
        :param database_id: The database id of the created identifier.
        :param type: The type of the created identifier.
        :param titles: The titles of the created identifier.
        :param publisher: The publisher of the created identifier.
        :param creators: The creator(s) of the created identifier.
        :param publication_year: The publication year of the created identifier.
        :param descriptions: The description(s) of the created identifier. Optional.
        :param funders: The funders(s) of the created identifier. Optional.
        :param licenses: The license(s) of the created identifier. Optional.
        :param language: The language of the created identifier. Optional.
        :param query_id: The query id of the created identifier. Required when type=SUBSET, otherwise invalid. Optional.
        :param view_id: The view id of the created identifier. Required when type=VIEW, otherwise invalid. Optional.
        :param table_id: The table id of the created identifier. Required when type=TABLE, otherwise invalid. Optional.
        :param publication_day: The publication day of the created identifier. Optional.
        :param publication_month: The publication month of the created identifier. Optional.
        :param related_identifiers: The related identifier(s) of the created identifier. Optional.

        :returns: The identifier, if successful.

        :raises ResponseCodeError: If something went wrong with the creation of the identifier.
        :raises ForbiddenError: If the action is not allowed.
        :raises MalformedError: If the payload is rejected by the service.
        :raises NotExistsError: If the database, table/view/subset or user does not exist.
        :raises ExternalSystemError: If the external system (DataCite) refused communication with the service.
        """
        url = f'/api/identifier/{identifier_id}'
        payload = CreateIdentifier(database_id=database_id, type=type, titles=titles, publisher=publisher,
                                   creators=creators, publication_year=publication_year, descriptions=descriptions,
                                   funders=funders, licenses=licenses, language=language, query_id=query_id,
                                   view_id=view_id, table_id=table_id, publication_day=publication_day,
                                   publication_month=publication_month, related_identifiers=related_identifiers)
        response = self._wrapper(method="put", url=url, force_auth=True, payload=payload)
        if response.status_code == 201:
            body = response.json()
            return Identifier.model_validate(body)
        if response.status_code == 400:
            raise MalformedError(f'Failed to save identifier: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to save identifier: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to save identifier: not found')
        if response.status_code == 503:
            raise ExternalSystemError(f'Failed to save identifier: external system rejected communication')
        raise ResponseCodeError(
            f'Failed to save identifier: response code: {response.status_code} is not 202 (ACCEPTED)')

    def publish_identifier(self, identifier_id: int) -> Identifier:
        """
        Publish an identifier with given id.

        :param identifier_id: The identifier id.

        :returns: The identifier, if successful.

        :raises ResponseCodeError: If something went wrong with the creation of the identifier.
        :raises ForbiddenError: If the action is not allowed.
        :raises MalformedError: If the payload is rejected by the service.
        :raises NotExistsError: If the database, table/view/subset or user does not exist.
        :raises ExternalSystemError: If the external system (DataCite) refused communication with the service.
        """
        url = f'/api/identifier/{identifier_id}/publish'
        response = self._wrapper(method="put", url=url, force_auth=True)
        if response.status_code == 201:
            body = response.json()
            return Identifier.model_validate(body)
        if response.status_code == 400:
            raise MalformedError(f'Failed to publish identifier: service rejected malformed payload')
        if response.status_code == 403 or response.status_code == 405:
            raise ForbiddenError(f'Failed to publish identifier: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to publish identifier: not found')
        if response.status_code == 503:
            raise ExternalSystemError(f'Failed to publish identifier: external system rejected communication')
        raise ResponseCodeError(
            f'Failed to publish identifier: response code: {response.status_code} is not 201 (CREATED)')

    def suggest_identifier(self, uri: str) -> Identifier:
        """
        Suggest identifier metadata for a given identifier URI. Example: ROR, ORCID, ISNI, GND, DOI.

        :param uri: The identifier URI.

        :returns: The identifier, if successful.

        :raises ResponseCodeError: If something went wrong with the suggestion of the identifier.
        :raises NotExistsError: If no metadata can be found or the identifier type is not supported.
        """
        url = f'/api/identifier?url={uri}'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return Identifier.model_validate(body)
        if response.status_code == 404:
            raise NotExistsError(f'Failed to suggest identifier: not found or not supported')
        raise ResponseCodeError(f'Failed to suggest identifier: response code: {response.status_code} is not 200 (OK)')

    def get_licenses(self) -> List[License]:
        """
        Get list of licenses allowed.

        :returns: List of licenses, if successful.
        """
        url = f'/api/database/license'
        response = self._wrapper(method="get", url=url)
        if response.status_code == 200:
            body = response.json()
            return TypeAdapter(List[License]).validate_python(body)
        raise ResponseCodeError(f'Failed to get licenses: response code: {response.status_code} is not 200 (OK)')

    def get_identifiers(self, ld: bool = False) -> List[Identifier] | str:
        """
        Get list of identifiers.

        :param ld: If set to true, identifiers are requested as JSON-LD. Optional. Default: false.

        :returns: List of identifiers, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval of the identifiers.
        :raises NotExistsError: If the accept header is neither application/json nor application/ld+json.
        """
        url = f'/api/pid'
        headers = None
        if ld:
            headers = {'Accept': 'application/ld+json'}
        response = self._wrapper(method="get", url=url, headers=headers)
        if response.status_code == 200:
            if ld:
                return response.json()
            else:
                body = response.json()
                return TypeAdapter(List[Identifier]).validate_python(body)
        if response.status_code == 406:
            raise MalformedError(
                f'Failed to get identifiers: accept header must be application/json or application/ld+json')
        raise ResponseCodeError(f'Failed to get identifiers: response code: {response.status_code} is not 200 (OK)')

    def update_table_column(self, database_id: int, table_id: int, column_id: int, concept_uri: str = None,
                            unit_uri: str = None) -> Column:
        """
        Update semantic information of a table column by given database id and table id and column id.

        :param database_id: The database id.
        :param table_id: The table id.
        :param column_id: The column id.
        :param concept_uri: The concept URI. Optional.
        :param unit_uri: The unit URI. Optional.

        :returns: The column, if successful.

        :raises ResponseCodeError: If something went wrong with the retrieval of the identifiers.
        :raises NotExistsError: If the accept header is neither application/json nor application/ld+json.
        """
        url = f'/api/database/{database_id}/table/{table_id}/column/{column_id}'
        response = self._wrapper(method="put", url=url, force_auth=True,
                                 payload=UpdateColumn(concept_uri=concept_uri, unit_uri=unit_uri))
        if response.status_code == 202:
            body = response.json()
            return Column.model_validate(body)
        if response.status_code == 400:
            raise MalformedError(f'Failed to update column: service rejected malformed payload')
        if response.status_code == 403:
            raise ForbiddenError(f'Failed to update colum: not allowed')
        if response.status_code == 404:
            raise NotExistsError(f'Failed to update colum: not found')
        raise ResponseCodeError(f'Failed to update colum: response code: {response.status_code} is not 202 (ACCEPTED)')



#!/usr/bin/env python3
from distutils.core import setup

setup(name="dbrepo",
      version="1.4.4rc2",
      description="A library for communicating with DBRepo",
      url="https://www.ifs.tuwien.ac.at/infrastructures/dbrepo/1.4.3/",
      author="Martin Weise",
      license="Apache-2.0",
      author_email="martin.weise@tuwien.ac.at",
      packages=[
            "dbrepo",
            "dbrepo.api"
      ])



import unittest
import requests_mock
import dataclasses

from dbrepo.RestClient import RestClient

from dbrepo.api.dto import KeyAnalysis


class AnalyseUnitTest(unittest.TestCase):

    def test_analyse_keys_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = KeyAnalysis(keys={'id': 0, 'firstname': 1, 'lastname': 2})
            # mock
            mock.get('/api/analyse/keys', json=exp.model_dump(), status_code=202)
            # test
            response = RestClient().analyse_keys(file_path='f705a7bd0cb2d5e37ab2b425036810a2', separator=',',
                                                 upload=False)
            self.assertEqual(exp, response)


if __name__ == "__main__":
    unittest.main()



import unittest

import requests_mock
import datetime

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Container, Image, ContainerBrief, ImageBrief
from dbrepo.api.exceptions import ResponseCodeError, NotExistsError

from dbrepo.api.dto import ImageDate


class ContainerUnitTest(unittest.TestCase):

    def test_get_containers_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/container', json=[])
            # test
            response = RestClient().get_containers()
            self.assertEqual([], response)

    def test_get_containers_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [
                ContainerBrief(id=1,
                               name="MariaDB 10.11.3",
                               internal_name="mariadb_10_11_3",
                               running=True,
                               created=datetime.datetime(2024, 3, 26, 10, 11, 0, 0, datetime.timezone.utc),
                               image=ImageBrief(id=1,
                                                name="mariadb",
                                                version="10.11.3",
                                                jdbc_method="mariadb"),
                               hash="f829dd8a884182d0da846f365dee1221fd16610a14c81b8f9f295ff162749e50")
            ]
            # mock
            mock.get('/api/container', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_containers()
            self.assertEqual(exp, response)

    def test_get_containers_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/container', status_code=204)
            # test
            try:
                response = RestClient().get_containers()
            except ResponseCodeError:
                pass

    def test_get_container_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Container(id=1,
                            name="MariaDB 10.11.3",
                            internal_name="mariadb_10_11_3",
                            running=True,
                            host="data-db",
                            port=12345,
                            sidecar_host="data-db-sidecar",
                            sidecar_port=3305,
                            created=datetime.datetime(2024, 3, 26, 10, 11, 0, 0, datetime.timezone.utc),
                            image=Image(id=1,
                                        registry="docker.io",
                                        name="mariadb",
                                        version="10.11.3",
                                        default_port=3306,
                                        dialect="org.hibernate.dialect.MariaDBDialect",
                                        driver_class="org.mariadb.jdbc.Driver",
                                        jdbc_method="mariadb",
                                        date_formats=[
                                            ImageDate(id=1,
                                                      example="2024-03-26 10:26:00",
                                                      database_format="%Y-%c-%d %H:%i:%S",
                                                      unix_format="yyyy-MM-dd HH:mm:ss",
                                                      has_time=True,
                                                      created_at=datetime.datetime(2024, 3, 26, 10, 26, 0, 0,
                                                                                   datetime.timezone.utc)),
                                            ImageDate(id=2,
                                                      example="2024-03-26",
                                                      database_format="%Y-%c-%d",
                                                      unix_format="yyyy-MM-dd",
                                                      has_time=False,
                                                      created_at=datetime.datetime(2024, 3, 26, 0, 0, 0, 0,
                                                                                   datetime.timezone.utc)),
                                            ImageDate(id=3,
                                                      example="10:25:01",
                                                      database_format="%Y-%c-%d",
                                                      unix_format="yyyy-MM-dd",
                                                      has_time=False,
                                                      created_at=datetime.datetime(2024, 3, 26, 0, 0, 0, 0,
                                                                                   datetime.timezone.utc)),
                                        ]),
                            hash="f829dd8a884182d0da846f365dee1221fd16610a14c81b8f9f295ff162749e50")
            # mock
            mock.get('/api/container/1', json=exp.model_dump())
            # test
            response = RestClient().get_container(container_id=1)
            self.assertEqual(exp, response)

    def test_get_container_not_found_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/container/1', status_code=404)
            # test
            try:
                response = RestClient().get_container(container_id=1)
            except NotExistsError:
                pass


if __name__ == "__main__":
    unittest.main()
